import logging
from django.db import models
from django.db.models.manager import Manager
from django.db.models.constants import LOOKUP_SEP

from rest_framework import serializers

import oscar.models.fields

from oscarapi.utils.exists import construct_id_filter
from .fields import ImageUrlField

logger = logging.getLogger(__name__)


def expand_field_mapping(extra_fields):
    # This doesn't make a copy
    field_mapping = serializers.ModelSerializer.serializer_field_mapping
    field_mapping.update(extra_fields)
    return field_mapping


class OscarSerializer(object):
    field_mapping = expand_field_mapping(
        {
            oscar.models.fields.NullCharField: serializers.CharField,
            models.ImageField: ImageUrlField,
        }
    )

    def __init__(self, *args, **kwargs):
        """
        Allow the serializer to be initiated with only a subset of the
        specified fields
        """
        fields = kwargs.pop("fields", None)
        super(OscarSerializer, self).__init__(*args, **kwargs)
        if fields:
            allowed = set(fields)
            existing = set(self.fields.keys())  # pylint: disable=no-member
            for field_name in existing - allowed:
                self.fields.pop(field_name)  # pylint: disable=no-member

    def update_relation(self, name, manager, values):
        if values is None:
            return

        serializer = self.fields[name]  # pylint: disable=no-member

        # use the serializer to update the attribute_values
        updated_values = serializer.update(manager, values)

        if self.partial:  # pylint: disable=no-member
            manager.add(*updated_values)
        elif hasattr(manager, "field") and not manager.field.null:
            # add the updated_attribute_values to the instance
            manager.add(*updated_values)
            # remove all the obsolete attribute values, this could be caused by
            # the product class changing for example, lots of attributes would become
            # obsolete.
            current_pks = [p.pk for p in updated_values]
            manager.exclude(pk__in=current_pks).delete()
        else:
            manager.set(updated_values)


class OscarModelSerializer(OscarSerializer, serializers.ModelSerializer):
    """
    Correctly map oscar fields to serializer fields.
    """


class OscarHyperlinkedModelSerializer(
    OscarSerializer, serializers.HyperlinkedModelSerializer
):
    """
    Correctly map oscar fields to serializer fields.
    """


class DelayUniqueSerializerMixin(object):
    def get_unique_together_validators(self):
        validators = super(
            DelayUniqueSerializerMixin, self
        ).get_unique_together_validators()

        if self.parent is not None:
            for validator in validators:
                validator.queryset = self.parent.restrict_query_to_parent(
                    validator.queryset
                )

        return validators


class UpdateListSerializer(serializers.ListSerializer):
    """
    This serializer can be used to implement updates to nested serializers.

    It will try to match existing objects based on the presence of a primary
    key in the submitted data, or the fields specified by name in ``Meta.lookup_fields``
    """

    class Meta:
        lookup_fields = []

    def restrict_query_to_parent(self, query):
        ParentModelClass = self.parent.Meta.model
        relation_field = ParentModelClass._meta.get_field(self.field_name)

        parent_field_name = relation_field.field.name

        if self.parent.instance:
            return query.exclude(**{parent_field_name: self.parent.instance})
        else:
            parent_filter = construct_id_filter(
                ParentModelClass,
                self.parent.get_initial(),
                prefix="%s%s" % (parent_field_name, LOOKUP_SEP),
            )
            return query.exclude(parent_filter).distinct()

    def select_existing_item(self, manager, datum):
        automatic_filter = construct_id_filter(manager.model, datum)

        intermediate_result = manager.filter(automatic_filter)
        try:
            return intermediate_result.distinct().get()
        except manager.model.DoesNotExist:
            pass
        except manager.model.MultipleObjectsReturned as e:
            logger.error("Multiple objects on unique constrained items, freaky %s", e)
            logger.exception(e)

        # fallback to manually specified lookup_fields
        for lookup_field in self.Meta.lookup_fields:
            if lookup_field in datum:
                try:
                    return intermediate_result.get(
                        **{lookup_field: datum[lookup_field]}
                    )
                except manager.model.DoesNotExist:
                    continue
                except manager.model.MultipleObjectsReturned:
                    continue

        return None

    def get_name_and_rel_instance(self, manager):
        return manager.field.name, manager.instance

    def update(self, instance, validated_data):
        assert isinstance(instance, Manager)

        field_name, rel_instance = self.get_name_and_rel_instance(instance)

        items = []
        for validated_datum in validated_data:
            complete_validated_datum = {field_name: rel_instance}
            complete_validated_datum.update(validated_datum)
            existing_item = self.select_existing_item(
                instance, complete_validated_datum
            )
            if existing_item is not None:
                updated_instance = self.child.update(
                    existing_item, complete_validated_datum
                )
            else:
                updated_instance = self.child.create(complete_validated_datum)

            items.append(updated_instance)

        return items


class UpdateForwardManyToManySerializer(UpdateListSerializer):
    def select_existing_item(self, manager, datum):
        # with manytomany there is no point in restricting the queryset
        # to the parent, because it is actually MEANT to share instances.
        return super(UpdateForwardManyToManySerializer, self).select_existing_item(
            manager.model.objects, datum
        )

    def get_name_and_rel_instance(self, manager):
        return manager.source_field_name, manager.instance



"""
This module contains functions that can be used to identify an existing piece
of data in the database based on it's unique attributes
"""
from django.db import models

from oscar.core.loading import get_model

from .accessors import destructure

AttributeOptionGroup = get_model("catalogue", "AttributeOptionGroup")
Category = get_model("catalogue", "Category")


def _field_name(name, prefix=None):
    """
    Util for quick prefixes

    >>> _field_name(1)
    '1'
    >>> _field_name("henk")
    'henk'
    >>> _field_name("henk", 1)
    '1henk'
    """
    if prefix is None:
        return str(name)
    return "%s%s" % (prefix, name)


def construct_id_filter(model, data, prefix=None):
    """
    This function will construct a filter that can be used to uniquely
    identify an object based on the keys present in the data object.

    So if there are multiple fields on a model that are marked as unique, or
    the model has unique_together specifucations, all these can be used to
    uniquely identify the instance that represents the data.
    """
    _filter = models.Q()

    for unique_together in model._meta.unique_together:
        _filter |= models.Q(
            **{_field_name(key, prefix): data.get(key) for key in unique_together}
        )

    for field in model._meta.concrete_fields:
        if field.unique and field.name in data and data[field.name] is not None:
            _filter |= models.Q(**{_field_name(field.name, prefix): data[field.name]})

    return _filter


def find_existing_attribute_option_group(name, options):
    query = (
        AttributeOptionGroup.objects.filter(name=name)
        .annotate(options_count=models.Count("options"))
        .filter(options_count=len(options))
    )
    for option in options:
        query = query.filter(options__option=option)

    try:
        return query.get()
    except (
        AttributeOptionGroup.DoesNotExist,
        AttributeOptionGroup.MultipleObjectsReturned,
    ):
        return None


def bound_unique_together_get_or_create(bound_queryset, datum):
    QuerySetModel = bound_queryset.model
    if isinstance(datum, QuerySetModel):
        return datum

    keys = set(datum.keys())
    for unique_pair in QuerySetModel._meta.unique_together:
        overlap = keys & set(unique_pair)
        # one of the items in the unique specification should be already
        # bound in the queryset (used in a filter). So we need to look
        # for a unique spec that has overlap with the data keys, but has 1
        # key missing (which is the bound key)
        if overlap and len(overlap) >= (len(unique_pair) - 1):
            unbound_keys = list(overlap)
            lookups = destructure(datum, *unbound_keys)
            item, _ = bound_queryset.get_or_create(**lookups)
            return item

    return None


def bound_unique_together_get_or_create_multiple(bound_queryset, data):
    return [bound_unique_together_get_or_create(bound_queryset, date) for date in data]



from rest_framework.fields import empty


def destructure(dictionary, *keys):
    """
    Select sub-dictionary

    >>> d = {"henk": 1, "klaas": 2, "hent": 7}
    >>> destructure(d, "klaas")
    {'klaas': 2}
    >>> sorted(destructure(d, "henk", "hent").items())
    [('henk', 1), ('hent', 7)]
    >>> destructure(d, "henk", "kak")
    {'henk': 1}
    """
    return dict(_getitems(dictionary, *keys))


def _getitems(dictionary, *keys):
    """
    Get existing items of a dict lazy

    >>> d = {"henk": 1, "klaas": 2, "hent": 7}
    >>> r = _getitems(d, "klaas")
    >>> list(r)
    [('klaas', 2)]
    >>> r = _getitems(d, "henk", "hent")
    >>> list(r)
    [('henk', 1), ('hent', 7)]
    >>> r = _getitems(d, "henk", "kak")
    >>> list(r)
    [('henk', 1)]
    """
    for key in keys:
        result = dictionary.get(key, empty)
        if result is empty:
            continue
        yield (key, result)


def getitems(dictionary, *keys):
    """
    Get existing items of a dict lazy

    >>> d = {"henk": 1, "klaas": 2, "hent": 7}
    >>> r = getitems(d, "klaas")
    >>> list(r)
    [2]
    >>> r = getitems(d, "henk", "hent")
    >>> list(r)
    [1, 7]
    >>> r = getitems(d, "henk", "kak")
    >>> list(r)
    [1, None]
    """
    for key in keys:
        yield dictionary.get(key)


def select(obj, *attrs):
    """
    Select attributes of object

    >>> from argparse import Namespace
    >>> d = Namespace(henk=1, klaas=2, hent=7)
    >>> select(d, "klaas")
    {'klaas': 2}
    >>> sorted(select(d, "henk", "hent").items())
    [('henk', 1), ('hent', 7)]
    >>> select(d, "henk", "kak")
    {'henk': 1}
    """
    return dict(_getattrs(obj, *attrs))


def _getattrs(obj, *attrs):
    """
    Get existing attributes of an object lazy

    >>> from argparse import Namespace
    >>> d = Namespace(henk=1, klaas=2, hent=7)
    >>> r = _getattrs(d, "klaas")
    >>> list(r)
    [('klaas', 2)]
    >>> r = _getattrs(d, "henk", "hent")
    >>> list(r)
    [('henk', 1), ('hent', 7)]
    >>> r = _getattrs(d, "henk", "kak")
    >>> list(r)
    [('henk', 1)]
    """
    for attr in attrs:
        result = getattr(obj, attr, empty)
        if result is empty:
            continue
        yield (attr, result)


def getattrs(obj, *attrs):
    """
    Get existing attributes of an object lazy

    >>> from argparse import Namespace
    >>> d = Namespace(henk=1, klaas=2, hent=7)
    >>> r = getattrs(d, "klaas")
    >>> list(r)
    [2]
    >>> r = getattrs(d, "henk", "hent")
    >>> list(r)
    [1, 7]
    >>> r = getattrs(d, "henk", "kak")
    >>> list(r)
    [1, None]
    """
    for attr in attrs:
        yield getattr(obj, attr, None)



from oscarapi.utils.loading import get_api_class
from oscar.core.loading import get_model
from rest_framework import generics

APIAdminPermission = get_api_class("permissions", "APIAdminPermission")
AdminStockRecordSerializer = get_api_class(
    "serializers.admin.partner", "AdminStockRecordSerializer"
)
PartnerSerializer = get_api_class("serializers.product", "PartnerSerializer")
StockRecord = get_model("partner", "StockRecord")
Partner = get_model("partner", "Partner")


class StockRecordDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = StockRecord.objects.all()
    serializer_class = AdminStockRecordSerializer
    permission_classes = (APIAdminPermission,)


class PartnerList(generics.ListCreateAPIView):
    queryset = Partner.objects.all()
    serializer_class = PartnerSerializer
    permission_classes = (APIAdminPermission,)


class PartnerDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = Partner.objects.all()
    serializer_class = PartnerSerializer
    permission_classes = (APIAdminPermission,)



import logging
import operator
import shutil
import tempfile

from os.path import basename, join, splitext
from urllib.parse import urlsplit, parse_qs
from urllib.request import Request, urlopen
from urllib.error import HTTPError
from django.conf import settings
from django.db import IntegrityError
from django.utils.translation import ugettext as _
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.files import File
from django.utils.functional import cached_property
from rest_framework import serializers, relations
from rest_framework.fields import get_attribute

from oscar.core.loading import get_model, get_class

from oscarapi import version
from oscarapi.utils.loading import get_api_class
from oscarapi.utils.exists import bound_unique_together_get_or_create
from oscarapi.utils.settings import overridable
from .exceptions import FieldError

logger = logging.getLogger(__name__)
ProductAttribute = get_model("catalogue", "ProductAttribute")
Category = get_model("catalogue", "Category")
create_from_breadcrumbs = get_class("catalogue.categories", "create_from_breadcrumbs")
entity_internal_value = get_api_class("serializers.hooks", "entity_internal_value")
attribute_details = operator.itemgetter("code", "value")


class TaxIncludedDecimalField(serializers.DecimalField):
    def __init__(self, excl_tax_field=None, excl_tax_value=None, **kwargs):
        self.excl_tax_field = excl_tax_field
        self.excl_tax_value = excl_tax_value
        super(TaxIncludedDecimalField, self).__init__(**kwargs)

    def get_attribute(self, instance):
        if instance.is_tax_known:
            return super(TaxIncludedDecimalField, self).get_attribute(instance)
        if self.excl_tax_field:
            return get_attribute(instance, (self.excl_tax_field,))
        return self.excl_tax_value


class DrillDownHyperlinkedMixin:
    def __init__(self, *args, **kwargs):
        try:
            self.extra_url_kwargs = kwargs.pop("extra_url_kwargs")
        except KeyError:
            msg = "DrillDownHyperlink Fields require an 'extra_url_kwargs' argument"
            raise ValueError(msg)

        super().__init__(*args, **kwargs)

    def get_extra_url_kwargs(self, obj):
        return {
            key: operator.attrgetter(path)(obj)
            for key, path in self.extra_url_kwargs.items()
        }

    def get_url(
        self, obj, view_name, request, format
    ):  # pylint: disable=redefined-builtin
        """
        Given an object, return the URL that hyperlinks to the object.

        May raise a `NoReverseMatch` if the `view_name` and `lookup_field`
        attributes are not configured to correctly match the URL conf.
        """
        if hasattr(obj, "pk") and obj.pk in (None, ""):
            return None

        lookup_value = getattr(obj, self.lookup_field)
        kwargs = {self.lookup_url_kwarg: lookup_value}
        kwargs.update(self.get_extra_url_kwargs(obj))
        return self.reverse(view_name, kwargs=kwargs, request=request, format=format)


class DrillDownHyperlinkedIdentityField(
    DrillDownHyperlinkedMixin, relations.HyperlinkedIdentityField
):
    pass


class DrillDownHyperlinkedRelatedField(
    DrillDownHyperlinkedMixin, relations.HyperlinkedRelatedField
):
    def use_pk_only_optimization(self):
        # we always want the full object so the mixin can filter on the attributes
        # specified with get_extra_url_kwargs
        return False


class AttributeValueField(serializers.Field):
    """
    This field is used to handle the value of the ProductAttributeValue model

    Because the value is dependant on the type of the corresponding attribute,
    it is not fixed. This field solves the problem of handling the different
    types.
    """

    def __init__(self, **kwargs):
        # this field always needs the full object
        kwargs["source"] = "*"
        kwargs["error_messages"] = {
            "no_such_option": _("{code}: Option {value} does not exist."),
            "invalid": _("Wrong type, {error}."),
            "attribute_validation_error": _(
                "Error assigning `{value}` to {code}, {error}."
            ),
            "attribute_required": _("Attribute {code} is required."),
            "attribute_missing": _(
                "No attribute exist with code={code}, "
                "please define it in the product_class first."
            ),
            "child_without_parent": _(
                "Can not find attribute if product_class is empty and "
                "parent is empty as well, child without parent?"
            ),
        }
        super(AttributeValueField, self).__init__(**kwargs)

    def get_value(self, dictionary):
        # return all the data because this field uses everything
        return dictionary

    def to_internal_value(self, data):  # noqa
        assert "product" in data or "product_class" in data or "parent" in data

        try:
            code, value = attribute_details(data)
            internal_value = value

            if "product" in data:
                # we need the attribute to determine the type of the value
                attribute = ProductAttribute.objects.get(
                    code=code, product_class__products__id=data["product"]
                )
            elif "product_class" in data and data["product_class"] is not None:
                attribute = ProductAttribute.objects.get(
                    code=code, product_class__slug=data.get("product_class")
                )
            elif "parent" in data:
                attribute = ProductAttribute.objects.get(
                    code=code, product_class__products__id=data["parent"]
                )

            if attribute.required and value is None:
                self.fail("attribute_required", code=code)

            # some of these attribute types need special processing, or their
            # validation will fail
            if attribute.type == attribute.OPTION:
                internal_value = attribute.option_group.options.get(option=value)
            elif attribute.type == attribute.MULTI_OPTION:
                if attribute.required and not value:
                    self.fail("attribute_required", code=code)
                internal_value = attribute.option_group.options.filter(option__in=value)
                if len(value) != internal_value.count():
                    non_existing = set(value) - set(
                        internal_value.values_list("option", flat=True)
                    )
                    non_existing_as_error = ",".join(sorted(non_existing))
                    self.fail("no_such_option", value=non_existing_as_error, code=code)
            elif attribute.type == attribute.DATE:
                date_field = serializers.DateField()
                internal_value = date_field.to_internal_value(value)
            elif attribute.type == attribute.DATETIME:
                date_field = serializers.DateTimeField()
                internal_value = date_field.to_internal_value(value)
            elif attribute.type == attribute.ENTITY:
                internal_value = entity_internal_value(attribute, value)

            # the rest of the attribute types don't need special processing
            try:
                attribute.validate_value(internal_value)
            except TypeError as e:
                self.fail(
                    "attribute_validation_error",
                    code=code,
                    value=internal_value,
                    error=e,
                )
            except ValidationError as e:
                self.fail(
                    "attribute_validation_error",
                    code=code,
                    value=internal_value,
                    error=",".join(e.messages),
                )

            return {"value": internal_value, "attribute": attribute}
        except ProductAttribute.DoesNotExist:
            if (
                "product_class" in data
                and "parent" in data
                and data["product_class"] is None
                and data["parent"] is None
            ):
                self.fail("child_without_parent")
            else:
                self.fail("attribute_missing", **data)
        except ObjectDoesNotExist:
            self.fail("no_such_option", value=value, code=code)
        except KeyError as e:
            (field_name,) = e.args
            raise FieldError(
                detail={field_name: self.error_messages["required"]}, code="required"
            )

    def to_representation(self, value):
        obj_type = value.attribute.type
        if obj_type == value.attribute.OPTION:
            return value.value.option
        elif obj_type == value.attribute.MULTI_OPTION:
            return value.value.values_list("option", flat=True)
        elif obj_type == value.attribute.FILE:
            return value.value.url
        elif obj_type == value.attribute.IMAGE:
            return value.value.url
        elif obj_type == value.attribute.ENTITY:
            if hasattr(value.value, "json"):
                return value.value.json()
            else:
                return _(
                    "%(entity)s has no json method, can not convert to json"
                    % {"entity": repr(value.value)}
                )

        # return the value as stored on ProductAttributeValue in the correct type
        return value.value


class CategoryField(serializers.RelatedField):
    def __init__(self, **kwargs):
        kwargs["queryset"] = Category.objects
        super(CategoryField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        return create_from_breadcrumbs(data)

    def to_representation(self, value):
        return value.full_name


class SingleValueSlugRelatedField(serializers.SlugRelatedField):
    """
    Represents a queryset as a list of slugs, and can be used to create new
    items, as long as only the slug_field is required
    """

    def get_bound_queryset(self):
        parent = self.parent
        source_name = parent.source
        if hasattr(parent, "child_relation"):
            parent = parent.parent

        return getattr(parent.instance, source_name, None)

    def to_internal_value(self, data):
        qs = self.get_bound_queryset()
        if qs is not None:  # first try to obtain a bound item.
            try:
                return bound_unique_together_get_or_create(qs, {self.slug_field: data})
            except IntegrityError:
                pass

        # if no bound item can be found, return an unbound unsaved instance.
        qs = self.get_queryset()
        return {self.slug_field: data}


class LazyRemoteFile(File):
    """
    This file will defer downloading untill the file data is accessed.

    It will also try to parsed a sha1 hash from the url, and store it as an
    attribute, so the file_hash function will use it. You can use this feature
    to avoid unnescessary downloading of files. Just compute the hash on the
    client side and send it along in the url like this::

        http://example.com/image.jpg?sha1=751499a82438277cb3cfb5db268bd41696739b3b

    It will only download if not allready available locally.
    """

    def __init__(self, url, name=None, mode="rb"):
        parsed_url = urlsplit(url)
        self.mode = mode
        self.name = name
        self.size = 1
        self.url = url

        # compute a hash if available
        sha1_hash = next(iter(parse_qs(parsed_url.query).get("sha1", [])), None)
        if sha1_hash:
            self.sha1 = sha1_hash

    def read(self, size=-1):
        try:
            return self.file.read(size)
        except HTTPError as e:
            raise serializers.ValidationError(
                "Error when downloading image %s, %s: %s" % (self.url, e.code, e.reason)
            )

    @cached_property
    def file(self):
        headers = overridable(
            "OSCARAPI_LAZY_REMOTE_FILE_REQUEST_HEADERS",
            default={"User-Agent": f"django-osscar-api/{version}"},
        )
        request = Request(self.url, headers=headers)

        _, ext = splitext(self.name)

        # Try to keep file in memory, but do not exceed FILE_UPLOAD_MAX_MEMORY_SIZE
        result_file = tempfile.SpooledTemporaryFile(
            max_size=settings.FILE_UPLOAD_MAX_MEMORY_SIZE,
            mode="wb",
            suffix=".upload" + ext,
            dir=settings.FILE_UPLOAD_TEMP_DIR,
        )
        response = urlopen(request)
        # copy the response into a file so it can be read multiple times
        shutil.copyfileobj(response, result_file)

        result_file.seek(0)
        return result_file

    def __str__(self):
        return self.url or ""

    def __bool__(self):
        return bool(self.url)

    def open(self, mode="rb"):
        if not self.closed:
            self.seek(0)

        return self


class ImageUrlField(serializers.ImageField):
    def __init__(self, **kwargs):
        super(ImageUrlField, self).__init__(**kwargs)
        self.use_url = True

    def to_internal_value(self, data):
        http_prefix = data.startswith(("http:", "https:"))
        if http_prefix:
            request = self.context.get("request", None)
            if request:  # if there is a request, we can get the hostname from that
                parsed_url = urlsplit(data)
                host = request.get_host()
                if (
                    host != parsed_url.netloc
                ):  # we are only downloading files from a foreign server
                    # download only when needed
                    return LazyRemoteFile(data, name=basename(parsed_url.path))
                else:
                    location = parsed_url.path
                    path = join(
                        settings.MEDIA_ROOT, location.replace(settings.MEDIA_URL, "", 1)
                    )
                    file_object = File(open(path, "rb"), name=basename(parsed_url.path))

                return super(ImageUrlField, self).to_internal_value(file_object)

        return super(ImageUrlField, self).to_internal_value(data)



from django.conf import settings

from rest_framework.permissions import (
    BasePermission,
    IsAuthenticated,
    DjangoModelPermissions,
)

from oscarapi.basket.operations import request_allows_access_to


class IsOwner(IsAuthenticated):
    """
    Permission that checks if this object has a foreign key pointing to the
    authenticated user of this request
    """

    def has_object_permission(self, request, view, obj):
        return obj.user == request.user


class APIAdminPermission(DjangoModelPermissions):
    """
    The permission for all the admin api views. You only get admin api access when:
    - OSCARAPI_BLOCK_ADMIN_API_ACCESS is false
    - you are a staff user (is_staff)
    - you have any of the model permissions needed (view / add / change / delete)

    Feel free to customize!
    """

    perms_map = {
        "GET": ["%(app_label)s.view_%(model_name)s"],
        "OPTIONS": ["%(app_label)s.view_%(model_name)s"],
        "HEAD": ["%(app_label)s.view_%(model_name)s"],
        "POST": ["%(app_label)s.add_%(model_name)s"],
        "PUT": ["%(app_label)s.change_%(model_name)s"],
        "PATCH": ["%(app_label)s.change_%(model_name)s"],
        "DELETE": ["%(app_label)s.delete_%(model_name)s"],
    }

    @staticmethod
    def disallowed_by_setting_and_request(request):
        return (
            getattr(settings, "OSCARAPI_BLOCK_ADMIN_API_ACCESS", True)
            or not request.user.is_staff
        )

    def has_permission(self, request, view):
        if self.disallowed_by_setting_and_request(request):
            return False
        return super(APIAdminPermission, self).has_permission(request, view)


class RequestAllowsAccessTo(BasePermission):
    def has_object_permission(self, request, view, obj):
        return request_allows_access_to(request, obj)



# pylint: disable=unbalanced-tuple-unpacking
from django.http import Http404
from rest_framework import generics
from rest_framework.exceptions import NotFound

from oscar.core.loading import get_model
from oscarapi.utils.loading import get_api_classes, get_api_class
from oscarapi.utils.exists import construct_id_filter

APIAdminPermission = get_api_class("permissions", "APIAdminPermission")
ProductAttributeSerializer, AttributeOptionGroupSerializer = get_api_classes(
    "serializers.product",
    ["ProductAttributeSerializer", "AttributeOptionGroupSerializer"],
)
(
    AdminProductSerializer,
    AdminCategorySerializer,
    AdminProductClassSerializer,
) = get_api_classes(
    "serializers.admin.product",
    [
        "AdminProductSerializer",
        "AdminCategorySerializer",
        "AdminProductClassSerializer",
    ],
)
CategoryList = get_api_class("views.product", "CategoryList")
Product = get_model("catalogue", "Product")
Category = get_model("catalogue", "Category")
ProductAttribute = get_model("catalogue", "ProductAttribute")
ProductClass = get_model("catalogue", "ProductClass")
AttributeOptionGroup = get_model("catalogue", "AttributeOptionGroup")


class ProductAdminList(generics.UpdateAPIView, generics.ListCreateAPIView):
    """
    Use this api for synchronizing data from another datasource.

    This api endpoint supports POST, PUT and PATCH, which means it can be used
    for creating, but also for updating. There is no need to supply the
    primary key(id) of the product, as long as you are sending enough data
    to uniquely identify the product (upc). That means you can try updating and
    if that fails, try POST. Or the other way around, whatever makes most
    sense in you scenario.

    Note that if you have changed the product model and changed upc to no longer
    be unique, you MUST add another unique field or specify a unique together
    constraint. And you have to send that data along.
    """

    serializer_class = AdminProductSerializer
    queryset = Product.objects.get_queryset()
    permission_classes = (APIAdminPermission,)

    def get_object(self):
        """
        Returns the object the view is displaying.

        Tries to extract a uniquely identifying query from the posted data
        """
        try:
            automatic_filter = construct_id_filter(Product, self.request.data)
            if automatic_filter:
                obj = Product.objects.get(automatic_filter)
                self.check_object_permissions(self.request, obj)
                return obj
            else:
                raise NotFound(
                    "Not enough info to identify %s." % Product._meta.object_name
                )
        except Product.DoesNotExist:
            raise Http404("No %s matches the given query." % Product._meta.object_name)


class ProductAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = AdminProductSerializer
    queryset = Product.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class ProductClassAdminList(generics.ListCreateAPIView):
    serializer_class = AdminProductClassSerializer
    queryset = ProductClass.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class ProductClassAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = AdminProductClassSerializer
    queryset = ProductClass.objects.get_queryset()
    permission_classes = (APIAdminPermission,)
    lookup_field = "slug"


class ProductAttributeAdminList(generics.ListCreateAPIView):
    serializer_class = ProductAttributeSerializer
    queryset = ProductAttribute.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class ProductAttributeAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = ProductAttributeSerializer
    queryset = ProductAttribute.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class AttributeOptionGroupAdminList(generics.ListCreateAPIView):
    serializer_class = AttributeOptionGroupSerializer
    queryset = AttributeOptionGroup.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class AttributeOptionGroupAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = AttributeOptionGroupSerializer
    queryset = AttributeOptionGroup.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class CategoryAdminList(generics.ListCreateAPIView, CategoryList):
    queryset = Category.get_root_nodes()
    serializer_class = AdminCategorySerializer
    permission_classes = (APIAdminPermission,)

    def get_queryset(self):
        try:
            return super(CategoryAdminList, self).get_queryset()
        except NotFound:  # admins might be able to create so hold the error.
            return Category.objects.none()

    def get_serializer_context(self):
        ctx = super(CategoryAdminList, self).get_serializer_context()
        breadcrumb_path = self.kwargs.get("breadcrumbs", None)

        if breadcrumb_path is not None:
            ctx["breadcrumbs"] = breadcrumb_path

        return ctx


class CategoryAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = Category.objects.all()
    serializer_class = AdminCategorySerializer
    permission_classes = (APIAdminPermission,)



import logging
from copy import deepcopy
from django.utils.translation import ugettext as _

from rest_framework import serializers
from rest_framework.fields import empty

from oscar.core.loading import get_model

from oscarapi.utils.exists import bound_unique_together_get_or_create_multiple
from oscarapi.utils.loading import get_api_classes
from oscarapi.utils.settings import overridable
from oscarapi.utils.files import file_hash
from oscarapi.utils.exists import find_existing_attribute_option_group
from oscarapi.utils.accessors import getitems

from oscarapi.serializers.fields import DrillDownHyperlinkedIdentityField
from oscarapi.serializers.utils import (
    OscarModelSerializer,
    OscarHyperlinkedModelSerializer,
    UpdateListSerializer,
    UpdateForwardManyToManySerializer,
)

from .exceptions import FieldError

logger = logging.getLogger(__name__)
Product = get_model("catalogue", "Product")
Range = get_model("offer", "Range")
ProductAttributeValue = get_model("catalogue", "ProductAttributeValue")
ProductImage = get_model("catalogue", "ProductImage")
Option = get_model("catalogue", "Option")
Partner = get_model("partner", "Partner")
StockRecord = get_model("partner", "StockRecord")
ProductClass = get_model("catalogue", "ProductClass")
ProductAttribute = get_model("catalogue", "ProductAttribute")
Category = get_model("catalogue", "Category")
AttributeOption = get_model("catalogue", "AttributeOption")
AttributeOptionGroup = get_model("catalogue", "AttributeOptionGroup")
AttributeValueField, CategoryField, SingleValueSlugRelatedField = get_api_classes(
    "serializers.fields",
    ["AttributeValueField", "CategoryField", "SingleValueSlugRelatedField"],
)


class AttributeOptionGroupSerializer(OscarHyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="admin-attributeoptiongroup-detail"
    )
    options = SingleValueSlugRelatedField(
        many=True,
        required=True,
        slug_field="option",
        queryset=AttributeOption.objects.get_queryset(),
    )

    def create(self, validated_data):
        existing_option = find_existing_attribute_option_group(
            validated_data["name"], validated_data["options"]
        )
        if existing_option is not None:
            return existing_option
        else:
            options = validated_data.pop("options", None)
            instance = super(AttributeOptionGroupSerializer, self).create(
                validated_data
            )
            options = bound_unique_together_get_or_create_multiple(
                instance.options, options
            )
            instance.options.set(options)
            return instance

    def update(self, instance, validated_data):
        existing_option = find_existing_attribute_option_group(
            validated_data["name"], validated_data["options"]
        )
        if existing_option is not None:
            return existing_option
        else:
            options = validated_data.pop("options", None)
            updated_instance = super(AttributeOptionGroupSerializer, self).update(
                instance, validated_data
            )
            # if the field was returned unbound,
            options = bound_unique_together_get_or_create_multiple(
                updated_instance.options, options
            )
            updated_instance.options.set(options, bulk=False)
            if not self.partial:
                # we need to manually remove the options
                updated_instance.options.exclude(
                    pk__in=[o.pk for o in options]
                ).delete()

            return updated_instance

    class Meta:
        model = AttributeOptionGroup
        fields = "__all__"


class BaseCategorySerializer(OscarHyperlinkedModelSerializer):
    breadcrumbs = serializers.CharField(source="full_name", read_only=True)

    class Meta:
        model = Category
        exclude = ("path", "depth", "numchild")


class CategorySerializer(BaseCategorySerializer):
    children = serializers.HyperlinkedIdentityField(
        view_name="category-child-list",
        lookup_field="full_slug",
        lookup_url_kwarg="breadcrumbs",
    )


class ProductAttributeListSerializer(UpdateListSerializer):
    def select_existing_item(self, manager, datum):
        try:
            return manager.get(product_class=datum["product_class"], code=datum["code"])
        except manager.model.DoesNotExist:
            pass
        except manager.model.MultipleObjectsReturned as e:
            logger.error("Multiple objects on unique contrained items, freaky %s", e)
            logger.exception(e)

        return None


class ProductAttributeSerializer(OscarHyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="admin-productattribute-detail"
    )
    product_class = serializers.SlugRelatedField(
        slug_field="slug",
        queryset=ProductClass.objects.get_queryset(),
        write_only=True,
        required=False,
    )
    option_group = AttributeOptionGroupSerializer(required=False, allow_null=True)

    def create(self, validated_data):
        option_group = validated_data.pop("option_group", None)
        instance = super(ProductAttributeSerializer, self).create(validated_data)
        return self.update(instance, {"option_group": option_group})

    def update(self, instance, validated_data):
        option_group = validated_data.pop("option_group", None)
        updated_instance = super(ProductAttributeSerializer, self).update(
            instance, validated_data
        )
        if option_group is not None:
            serializer = self.fields["option_group"]
            # use the serializer to update the attribute_values
            if instance.option_group:
                updated_instance.option_group = serializer.update(
                    instance.option_group, option_group
                )
            else:
                updated_instance.option_group = serializer.create(option_group)

            updated_instance.save()

        return updated_instance

    class Meta:
        model = ProductAttribute
        list_serializer_class = ProductAttributeListSerializer
        fields = "__all__"


class RangeSerializer(OscarHyperlinkedModelSerializer):
    class Meta:
        model = Range
        fields = "__all__"


class PartnerSerializer(OscarHyperlinkedModelSerializer):
    class Meta:
        model = Partner
        fields = "__all__"


class OptionSerializer(OscarHyperlinkedModelSerializer):
    code = serializers.SlugField()

    class Meta:
        model = Option
        fields = overridable("OSCARAPI_OPTION_FIELDS", default="__all__")
        list_serializer_class = UpdateForwardManyToManySerializer


class ProductAttributeValueListSerializer(UpdateListSerializer):
    def get_value(self, dictionary):
        values = super(ProductAttributeValueListSerializer, self).get_value(dictionary)
        if values is empty:
            return values

        product_class, parent = getitems(dictionary, "product_class", "parent")
        return [
            dict(value, product_class=product_class, parent=parent) for value in values
        ]


class ProductAttributeValueSerializer(OscarModelSerializer):
    # we declare the product as write_only since this serializer is meant to be
    # used nested inside a product serializer.
    product = serializers.PrimaryKeyRelatedField(
        many=False, write_only=True, required=False, queryset=Product.objects
    )

    value = AttributeValueField()  # handles different attribute value types
    # while code is specified as read_only, it is still required, because otherwise
    # the attribute is unknown, so while it will never be overwritten, you do
    # have to include it in your data structure
    code = serializers.CharField(source="attribute.code", read_only=True)
    name = serializers.CharField(
        source="attribute.name", required=False, read_only=True
    )

    def to_internal_value(self, data):
        try:
            internal_value = super(
                ProductAttributeValueSerializer, self
            ).to_internal_value(data)
            internal_value["product_class"] = data.get("product_class")
            return internal_value
        except FieldError as e:
            raise serializers.ValidationError(e.detail)

    def save(self, **kwargs):
        """
        Since there is a unique constraint, sometimes we want to update instead
        of creating a new object (because an integrity error would occur due
        to the constraint on attribute and product). If instance is set, the
        update method will be used instead of the create method.
        """
        data = deepcopy(kwargs)
        data.update(self.validated_data)
        return self.update_or_create(data)

    def update_or_create(self, validated_data):
        value = validated_data["value"]
        product = validated_data["product"]
        attribute = validated_data["attribute"]
        attribute.save_value(product, value)
        return product.attribute_values.get(attribute=attribute)

    create = update_or_create

    def update(self, instance, validated_data):
        data = deepcopy(validated_data)
        data["product"] = instance.product
        return self.update_or_create(data)

    class Meta:
        model = ProductAttributeValue
        list_serializer_class = ProductAttributeValueListSerializer
        fields = overridable(
            "OSCARAPI_PRODUCT_ATTRIBUTE_VALUE_FIELDS",
            default=("name", "value", "code", "product"),
        )


class ProductImageUpdateListSerializer(UpdateListSerializer):
    "Select existing image based on hash of image content"

    def select_existing_item(self, manager, datum):
        # determine the hash of the passed image
        target_file_hash = file_hash(datum["original"])
        for image in manager.all():  # search for a match in the set of exising images
            _hash = file_hash(image.original)
            if _hash == target_file_hash:
                # django will create a copy of the original under a weird name,
                # because the image is freshly fetched, except if we use the
                # original image FileObject
                datum["original"] = image.original
                return image

        return None


class ProductImageSerializer(OscarModelSerializer):
    product = serializers.PrimaryKeyRelatedField(
        write_only=True, required=False, queryset=Product.objects
    )

    class Meta:
        model = ProductImage
        fields = "__all__"
        list_serializer_class = ProductImageUpdateListSerializer


class AvailabilitySerializer(serializers.Serializer):  # pylint: disable=abstract-method
    is_available_to_buy = serializers.BooleanField()
    num_available = serializers.IntegerField(required=False)
    message = serializers.CharField()


class RecommmendedProductSerializer(OscarModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="product-detail")

    class Meta:
        model = Product
        fields = overridable("OSCARAPI_RECOMMENDED_PRODUCT_FIELDS", default=("url",))


class ProductStockRecordSerializer(OscarModelSerializer):
    url = DrillDownHyperlinkedIdentityField(
        view_name="product-stockrecord-detail",
        extra_url_kwargs={"product_pk": "product_id"},
    )

    class Meta:
        model = StockRecord
        fields = "__all__"


class BaseProductSerializer(OscarModelSerializer):
    "Base class shared by admin and public serializer"
    attributes = ProductAttributeValueSerializer(
        many=True, required=False, source="attribute_values"
    )
    categories = CategoryField(many=True, required=False)
    product_class = serializers.SlugRelatedField(
        slug_field="slug", queryset=ProductClass.objects, allow_null=True
    )
    options = OptionSerializer(many=True, required=False)
    recommended_products = serializers.HyperlinkedRelatedField(
        view_name="product-detail",
        many=True,
        required=False,
        queryset=Product.objects.filter(
            structure__in=[Product.PARENT, Product.STANDALONE]
        ),
    )

    def validate(self, attrs):
        if "structure" in attrs and "parent" in attrs:
            if attrs["structure"] == Product.CHILD and attrs["parent"] is None:
                raise serializers.ValidationError(_("child without parent"))
        if "structure" in attrs and "product_class" in attrs:
            if attrs["product_class"] is None and attrs["structure"] != Product.CHILD:
                raise serializers.ValidationError(
                    _("product_class can not be empty for structure %(structure)s")
                    % attrs
                )

        return super(BaseProductSerializer, self).validate(attrs)

    class Meta:
        model = Product


class PublicProductSerializer(BaseProductSerializer):
    "Serializer base class used for public products api"
    url = serializers.HyperlinkedIdentityField(view_name="product-detail")
    price = serializers.HyperlinkedIdentityField(
        view_name="product-price", read_only=True
    )
    availability = serializers.HyperlinkedIdentityField(
        view_name="product-availability", read_only=True
    )

    def get_field_names(self, declared_fields, info):
        """
        Override get_field_names to make sure that we are not getting errors
        for not including declared fields.
        """
        return super(PublicProductSerializer, self).get_field_names({}, info)


class ChildProductSerializer(PublicProductSerializer):
    "Serializer for child products"
    parent = serializers.HyperlinkedRelatedField(
        view_name="product-detail",
        queryset=Product.objects.filter(structure=Product.PARENT),
    )
    # the below fields can be filled from the parent product if enabled.
    images = ProductImageSerializer(many=True, required=False, source="parent.images")
    description = serializers.CharField(source="parent.description")

    class Meta(PublicProductSerializer.Meta):
        fields = overridable(
            "OSCARAPI_CHILDPRODUCTDETAIL_FIELDS",
            default=(
                "url",
                "upc",
                "id",
                "title",
                "structure",
                # 'parent', 'description', 'images', are not included by default, but
                # easily enabled by overriding OSCARAPI_CHILDPRODUCTDETAIL_FIELDS
                # in your settings file
                "date_created",
                "date_updated",
                "recommended_products",
                "attributes",
                "categories",
                "product_class",
                "price",
                "availability",
                "options",
            ),
        )


class ProductSerializer(PublicProductSerializer):
    "Serializer for public api with strategy fields added for price and availability"
    url = serializers.HyperlinkedIdentityField(view_name="product-detail")
    price = serializers.HyperlinkedIdentityField(
        view_name="product-price", read_only=True
    )
    availability = serializers.HyperlinkedIdentityField(
        view_name="product-availability", read_only=True
    )

    images = ProductImageSerializer(many=True, required=False)
    children = ChildProductSerializer(many=True, required=False)

    stockrecords = serializers.HyperlinkedIdentityField(
        view_name="product-stockrecords", read_only=True
    )

    class Meta(PublicProductSerializer.Meta):
        fields = overridable(
            "OSCARAPI_PRODUCTDETAIL_FIELDS",
            default=(
                "url",
                "upc",
                "id",
                "title",
                "description",
                "structure",
                "date_created",
                "date_updated",
                "recommended_products",
                "attributes",
                "categories",
                "product_class",
                "images",
                "price",
                "availability",
                "stockrecords",
                "options",
                "children",
            ),
        )


class ProductLinkSerializer(ProductSerializer):
    """
    Summary serializer for list view, listing all products.

    This serializer can be easily made to show any field on ``ProductSerializer``,
    just add fields to the ``OSCARAPI_PRODUCT_FIELDS`` setting.
    """

    class Meta(PublicProductSerializer.Meta):
        fields = overridable(
            "OSCARAPI_PRODUCT_FIELDS", default=("url", "id", "upc", "title")
        )


class OptionValueSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    option = serializers.HyperlinkedRelatedField(
        view_name="option-detail", queryset=Option.objects
    )
    value = serializers.CharField()


class AddProductSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    """
    Serializes and validates an add to basket request.
    """

    quantity = serializers.IntegerField(required=True)
    url = serializers.HyperlinkedRelatedField(
        view_name="product-detail", queryset=Product.objects, required=True
    )
    options = OptionValueSerializer(many=True, required=False)



from django.utils.translation import ugettext as _

from rest_framework.exceptions import NotFound

from oscar.core.loading import get_model

Category = get_model("catalogue", "category")


def create_from_sequence(bits, create):
    """
    Create categories from an iterable
    """
    if len(bits) == 1:
        # Get or create root node
        slug = bits[0]
        try:
            # Category names should be unique at the depth=1
            root = Category.objects.get(depth=1, slug=slug)
        except Category.DoesNotExist:
            if create:
                root = Category.add_root(name=slug, slug=slug)
            else:
                raise NotFound(
                    _("No %(verbose_name)s found matching the query")
                    % {"verbose_name": Category._meta.verbose_name}
                )
        except Category.MultipleObjectsReturned:
            raise ValueError(
                "There are more than one categories with slug " "%s at depth=1" % slug
            )
        return [root]
    else:
        parents = create_from_sequence(bits[:-1], create)
        parent, slug = parents[-1], bits[-1]
        try:
            child = parent.get_children().get(slug=slug)
        except Category.DoesNotExist:
            if create:
                child = parent.add_child(name=slug, slug=slug)
            else:
                raise NotFound(
                    _("No %(verbose_name)s found matching the query")
                    % {"verbose_name": Category._meta.verbose_name}
                )
        except Category.MultipleObjectsReturned:
            raise ValueError(
                (
                    "There are more than one categories with slug "
                    "%s which are children of %s"
                )
                % (slug, parent)
            )
        parents.append(child)
        return parents


def create_from_full_slug(breadcrumb_str, separator="/"):
    """
    Create categories from a breadcrumb string
    """
    category_names = [x.strip() for x in breadcrumb_str.split(separator)]
    categories = create_from_sequence(category_names, True)
    return categories[-1]


def find_from_full_slug(breadcrumb_str, separator="/"):
    """
    Find categories from a breadcrumb string
    """
    category_names = [x.strip() for x in breadcrumb_str.split(separator)]
    categories = create_from_sequence(category_names, False)
    return categories[-1]



from rest_framework import serializers

from oscarapi.utils.loading import get_api_classes

OrderSerializer, OrderLineSerializer, OrderLineAttributeSerializer = get_api_classes(
    "serializers.checkout",
    ["OrderSerializer", "OrderLineSerializer", "OrderLineAttributeSerializer"],
)


class AdminOrderSerializer(OrderSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="admin-order-detail")
    lines = serializers.HyperlinkedIdentityField(view_name="admin-order-lines-list")


class AdminOrderLineAttributeSerializer(OrderLineAttributeSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="admin-order-lineattributes-detail"
    )


class AdminOrderLineSerializer(OrderLineSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="admin-order-lines-detail")
    attributes = AdminOrderLineAttributeSerializer(
        many=True, fields=("url", "option", "value"), required=False
    )



from django.contrib.auth import get_user_model

from rest_framework import serializers

from oscarapi.utils.settings import overridable

User = get_user_model()


class AdminUserSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="admin-user-detail")
    date_joined = serializers.DateTimeField(read_only=True)

    class Meta:
        model = User
        fields = overridable(
            "OSCARAPI_ADMIN_USER_FIELDS",
            default=("url", User.USERNAME_FIELD, "email", "date_joined"),
        )



from rest_framework import serializers

from oscar.core.loading import get_model
from oscarapi.serializers.utils import (
    DelayUniqueSerializerMixin,
    OscarHyperlinkedModelSerializer,
    UpdateListSerializer,
)

Product = get_model("catalogue", "Product")
StockRecord = get_model("partner", "StockRecord")


class AdminStockRecordSerializer(
    DelayUniqueSerializerMixin, OscarHyperlinkedModelSerializer
):
    url = serializers.HyperlinkedIdentityField(view_name="admin-stockrecord-detail")

    product = serializers.PrimaryKeyRelatedField(
        many=False, required=False, queryset=Product.objects
    )

    class Meta:
        model = StockRecord
        fields = "__all__"
        list_serializer_class = UpdateListSerializer



from django.db import transaction
from django.template.defaultfilters import slugify

from rest_framework import serializers
from rest_framework.exceptions import APIException

from oscar.core.loading import get_model

from oscarapi.serializers.utils import OscarHyperlinkedModelSerializer
from oscarapi.utils.categories import create_from_full_slug
from oscarapi.utils.loading import get_api_classes, get_api_class
from oscarapi.utils.models import fake_autocreated

Product = get_model("catalogue", "Product")
ProductClass = get_model("catalogue", "ProductClass")
ProductAttributeValue = get_model("catalogue", "ProductAttributeValue")
Option = get_model("catalogue", "Option")
AdminStockRecordSerializer = get_api_class(
    "serializers.admin.partner", "AdminStockRecordSerializer"
)
(
    BaseProductSerializer,
    BaseCategorySerializer,
    ProductImageSerializer,
    ProductAttributeSerializer,
    OptionSerializer,
) = get_api_classes(
    "serializers.product",
    [
        "BaseProductSerializer",
        "BaseCategorySerializer",
        "ProductImageSerializer",
        "ProductAttributeSerializer",
        "OptionSerializer",
    ],
)


class AdminProductSerializer(BaseProductSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="admin-product-detail")
    stockrecords = AdminStockRecordSerializer(many=True, required=False)
    images = ProductImageSerializer(many=True, required=False)
    children = serializers.HyperlinkedRelatedField(
        view_name="admin-product-detail",
        many=True,
        required=False,
        queryset=Product.objects.filter(structure=Product.CHILD),
    )

    class Meta(BaseProductSerializer.Meta):
        exclude = ("product_options",)

    def create(self, validated_data):
        attribute_values = validated_data.pop("attribute_values", None)
        options = validated_data.pop("options", None)
        stockrecords = validated_data.pop("stockrecords", None)
        images = validated_data.pop("images", None)
        categories = validated_data.pop("categories", None)
        recommended_products = validated_data.pop("recommended_products", None)
        children = validated_data.pop("children", None)

        with transaction.atomic():  # it is all or nothing!

            # update instance
            self.instance = (  # pylint:disable=attribute-defined-outside-init
                instance
            ) = super(AdminProductSerializer, self).create(validated_data)
            return self.update(
                instance,
                dict(
                    validated_data,
                    attribute_values=attribute_values,
                    options=options,
                    stockrecords=stockrecords,
                    images=images,
                    categories=categories,
                    recommended_products=recommended_products,
                    children=children,
                ),
            )

    def update(self, instance, validated_data):
        "Handle the nested serializers manually"
        attribute_values = validated_data.pop("attribute_values", None)
        options = validated_data.pop("options", None)
        stockrecords = validated_data.pop("stockrecords", None)
        images = validated_data.pop("images", None)
        categories = validated_data.pop("categories", None)
        recommended_products = validated_data.pop("recommended_products", None)
        children = validated_data.pop("children", None)

        with transaction.atomic():  # it is all or nothing!

            # update instance
            instance = super(AdminProductSerializer, self).update(
                instance, validated_data
            )

            # ``fake_autocreated`` removes the very annoying "Cannot set values
            # on a ManyToManyField which specifies an intermediary model" error,
            # which does not apply at all to these models because they have sane
            # defaults.

            if categories is not None:
                with fake_autocreated(instance.categories) as _categories:
                    if self.partial:
                        _categories.add(*categories)
                    else:
                        _categories.set(categories)

            if recommended_products is not None:
                with fake_autocreated(
                    instance.recommended_products
                ) as _recommended_products:
                    if self.partial:
                        _recommended_products.add(*recommended_products)
                    else:
                        _recommended_products.set(recommended_products)

            if children is not None:
                if self.partial:
                    instance.children.add(*children)
                else:
                    instance.children.set(children)

            product_class = instance.get_product_class()
            pclass_option_codes = set()
            if options is not None:
                with fake_autocreated(instance.product_options) as _product_options:
                    pclass_option_codes = product_class.options.filter(
                        code__in=[opt["code"] for opt in options]
                    ).values_list("code", flat=True)
                    # only options not allready defined on the product class are important
                    new_options = [
                        opt for opt in options if opt["code"] not in pclass_option_codes
                    ]
                    self.update_relation("options", _product_options, new_options)

            self.update_relation("images", instance.images, images)
            self.update_relation("stockrecords", instance.stockrecords, stockrecords)
            self.update_relation(
                "attributes", instance.attribute_values, attribute_values
            )

            if (
                self.partial
            ):  # we need to clean up all the attributes with wrong product class
                for attribute_value in instance.attribute_values.exclude(
                    attribute__product_class=product_class
                ):
                    code = attribute_value.attribute.code
                    if (
                        code in pclass_option_codes
                    ):  # if the attribute exist also on the new product class, update the attribute
                        attribute_value.attribute = product_class.attributes.get(
                            code=code
                        )
                        attribute_value.save()
                    else:
                        attribute_value.delete()
            # return a refreshed instance so we are sure all attributes are reloaded
            # from the database again when accessed. An attr.refresh() method
            # was added in Oscar 3, but that will not delete any existsing
            # attributes which  have become invalid (after updating the
            # product class for example)
            return instance._meta.model.objects.get(pk=instance.pk)


class AdminCategorySerializer(BaseCategorySerializer):
    url = serializers.HyperlinkedIdentityField(view_name="admin-category-detail")
    children = serializers.HyperlinkedIdentityField(
        view_name="admin-category-child-list",
        lookup_field="full_slug",
        lookup_url_kwarg="breadcrumbs",
    )
    slug = serializers.SlugField(required=False)

    def create(self, validated_data):
        breadcrumbs = self.context.get("breadcrumbs", None)
        slug = validated_data.get("slug", slugify(validated_data["name"]))

        if breadcrumbs is None:
            breadcrumbs = slug
        else:
            breadcrumbs = "/".join((breadcrumbs, slug))

        try:
            instance = create_from_full_slug(breadcrumbs, separator="/")
        except ValueError as e:
            raise APIException(str(e))

        return self.update(instance, validated_data)


class AdminProductClassSerializer(OscarHyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="admin-productclass-detail", lookup_field="slug"
    )
    attributes = ProductAttributeSerializer(many=True, required=False)
    options = OptionSerializer(many=True, required=False)

    def create(self, validated_data):
        attributes = validated_data.pop("attributes", None)
        options = validated_data.pop("options", None)

        with transaction.atomic():
            self.instance = instance = super(AdminProductClassSerializer, self).create(
                validated_data
            )
            return self.update(
                instance, dict(validated_data, attributes=attributes, options=options)
            )

    def update(self, instance, validated_data):
        attributes = validated_data.pop("attributes", None)
        options = validated_data.pop("options", None)

        with transaction.atomic():
            updated_instance = super(AdminProductClassSerializer, self).update(
                instance, validated_data
            )
            self.update_relation("attributes", updated_instance.attributes, attributes)
            self.update_relation("options", instance.options, options)

            return updated_instance

    class Meta:
        model = ProductClass
        fields = "__all__"



from contextlib import contextmanager

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


@contextmanager
def fake_autocreated(many_to_many_manager):
    "Do not give a shit about any intermediate models, just update the relation"
    with patch.object(many_to_many_manager.through._meta, "auto_created", True):
        yield many_to_many_manager



# pylint: disable=unbalanced-tuple-unpacking
from rest_framework import generics

from oscar.core.loading import get_model
from oscarapi.utils.loading import get_api_classes, get_api_class


APIAdminPermission = get_api_class("permissions", "APIAdminPermission")
Order = get_model("order", "Order")
OrderLine = get_model("order", "Line")
OrderLineAttribute = get_model("order", "LineAttribute")

(
    AdminOrderSerializer,
    AdminOrderLineSerializer,
    AdminOrderLineAttributeSerializer,
) = get_api_classes(  # noqa
    "serializers.admin.order",
    [
        "AdminOrderSerializer",
        "AdminOrderLineSerializer",
        "AdminOrderLineAttributeSerializer",
    ],
)


class OrderAdminList(generics.ListAPIView):
    serializer_class = AdminOrderSerializer
    queryset = Order.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class OrderAdminDetail(generics.RetrieveDestroyAPIView):
    serializer_class = AdminOrderSerializer
    queryset = Order.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class OrderLineAdminList(generics.ListAPIView):
    serializer_class = AdminOrderLineSerializer
    queryset = OrderLine.objects.get_queryset()
    permission_classes = (APIAdminPermission,)

    def get_queryset(self):
        pk = self.kwargs.get("pk")
        return super().get_queryset().filter(order_id=pk)


class OrderLineAdminDetail(generics.RetrieveDestroyAPIView):
    serializer_class = AdminOrderLineSerializer
    queryset = OrderLine.objects.get_queryset()
    permission_classes = (APIAdminPermission,)


class OrderLineAttributeAdminDetail(generics.RetrieveDestroyAPIView):
    serializer_class = AdminOrderLineAttributeSerializer
    queryset = OrderLineAttribute.objects.get_queryset()
    permission_classes = (APIAdminPermission,)



from django.core.exceptions import ValidationError

from oscar.core.loading import get_model

from oscarapi import permissions

from rest_framework import exceptions, generics
from rest_framework.relations import HyperlinkedRelatedField

__all__ = ("BasketPermissionMixin",)

Basket = get_model("basket", "Basket")


class QuerySetList(list):
    def __init__(self, some_list, queryset):
        super(QuerySetList, self).__init__(some_list)
        self.queryset = queryset

    @property
    def model(self):
        return self.queryset.model


def parse_basket_from_hyperlink(DATA, format):  # pylint: disable=redefined-builtin
    "Parse basket from relation hyperlink"
    basket_parser = HyperlinkedRelatedField(
        view_name="basket-detail", queryset=Basket.objects, format=format
    )
    try:
        basket_uri = DATA.get("basket")
        data_basket = basket_parser.to_internal_value(basket_uri)
    except ValidationError as e:
        raise exceptions.NotAcceptable(e.messages)
    else:
        return data_basket


class BasketPermissionMixin(object):
    """
    This mixins adds some methods that can be used to check permissions
    on a basket instance.
    """

    # The permission class is mainly used to check Basket permission!
    permission_classes = (permissions.RequestAllowsAccessTo,)

    def get_data_basket(self, DATA, format):  # pylint: disable=redefined-builtin
        return parse_basket_from_hyperlink(DATA, format)

    def check_basket_permission(self, request, basket_pk=None, basket=None):
        "Check if the user may access this basket"
        if basket is None:
            basket = generics.get_object_or_404(Basket.objects, pk=basket_pk)
        self.check_object_permissions(request, basket)
        return basket



import hashlib
from importlib import import_module

from django.conf import settings
from django.contrib import auth

from rest_framework import exceptions

from oscar.core.loading import get_class

Selector = get_class("partner.strategy", "Selector")


def login_and_upgrade_session(request, user):
    "Upgrade anonymous session to authenticated session"
    parsed_session_uri = getattr(request, "parsed_session_uri", None)

    if parsed_session_uri is not None:

        assert parsed_session_uri["type"] == "ANON"

        # change anonymous session to authenticated
        parsed_session_uri["type"] = "AUTH"
        session_id = session_id_from_parsed_session_uri(parsed_session_uri)

        # wipe out old anonymous session without creating a new one
        request.session.clear()
        request.session.delete()

        # start session with new session id
        request.session = get_session(session_id)

        # Mark the new session as owned by the user we are logging in.
        request.session[auth.SESSION_KEY] = user.pk
        if hasattr(user, "get_session_auth_hash"):  # django 1.7
            request.session[auth.HASH_SESSION_KEY] = user.get_session_auth_hash()

    # now login so the session can be used for authentication purposes.
    auth.login(request, user)
    request.session.save()


def session_id_from_parsed_session_uri(parsed_session_uri):
    session_id_base = "SID:%(type)s:%(realm)s:%(session_id)s" % (parsed_session_uri)
    combined = session_id_base + settings.SECRET_KEY
    return hashlib.sha1(combined.encode()).hexdigest()


def get_session(session_id, raise_on_create=False):
    "get a session with the id specified."
    engine = import_module(settings.SESSION_ENGINE)
    session = engine.SessionStore(session_id)

    if not session.exists(session_id):
        if raise_on_create:
            raise exceptions.NotAuthenticated()
        else:
            session.save(must_create=True)
    else:
        # if we get an expired session from django,
        # the session key will change after calling load.
        # since the whole point of get_session is to retrieve a session
        # with exactly the key specified, we have to clear the expired
        # sessions and then get a new session.
        session.load()
        if session.session_key != session_id:
            engine.SessionStore.clear_expired()
            session = get_session(session_id, raise_on_create)

    return session



from django.contrib.auth import get_user_model
from oscarapi.utils.loading import get_api_class
from rest_framework import generics


APIAdminPermission = get_api_class("permissions", "APIAdminPermission")
AdminUserSerializer = get_api_class("serializers.admin.user", "AdminUserSerializer")
User = get_user_model()


class UserAdminList(generics.ListCreateAPIView):
    """
    List of all users, either frontend or admin users.
    The fields shown in this view can be changed using the ``OSCARAPI_ADMIN_USER_FIELDS``
    setting
    """

    queryset = User.objects.all()
    serializer_class = AdminUserSerializer
    permission_classes = (APIAdminPermission,)


class UserAdminDetail(generics.RetrieveUpdateDestroyAPIView):
    queryset = User.objects.all()
    serializer_class = AdminUserSerializer
    permission_classes = (APIAdminPermission,)



from django.utils.translation import ugettext_lazy as _

from oscar.apps.basket import signals
from oscar.core.loading import get_model, get_class

from rest_framework import status, generics, exceptions
from rest_framework.response import Response
from rest_framework.views import APIView

from oscarapi import permissions
from oscarapi.basket import operations
from oscarapi.utils.loading import get_api_classes, get_api_class
from oscarapi.views.utils import BasketPermissionMixin

__all__ = (
    "BasketView",
    "LineList",
    "AddProductView",
    "BasketLineDetail",
    "AddVoucherView",
    "ShippingMethodView",
)

Basket = get_model("basket", "Basket")
Line = get_model("basket", "Line")
Repository = get_class("shipping.repository", "Repository")
ShippingAddress = get_model("order", "ShippingAddress")
(  # pylint: disable=unbalanced-tuple-unpacking
    BasketSerializer,
    VoucherAddSerializer,
    VoucherSerializer,
    BasketLineSerializer,
) = get_api_classes(
    "serializers.basket",
    [
        "BasketSerializer",
        "VoucherAddSerializer",
        "VoucherSerializer",
        "BasketLineSerializer",
    ],
)
AddProductSerializer = get_api_class("serializers.product", "AddProductSerializer")
(
    ShippingAddressSerializer,
    ShippingMethodSerializer,
) = get_api_classes(  # pylint: disable=unbalanced-tuple-unpacking
    "serializers.checkout", ["ShippingAddressSerializer", "ShippingMethodSerializer"]
)


class BasketView(APIView):
    """
    Api for retrieving a user's basket.

    GET:
    Retrieve your basket.
    """

    serializer_class = BasketSerializer

    def get(self, request, *args, **kwargs):  # pylint: disable=redefined-builtin
        basket = operations.get_basket(request)
        ser = self.serializer_class(basket, context={"request": request})
        return Response(ser.data)


class AddProductView(APIView):
    """
    Add a certain quantity of a product to the basket.

    POST(url, quantity)
    {
        "url": "http://testserver.org/oscarapi/products/209/",
        "quantity": 6
    }

    If you've got some options to configure for the product to add to the
    basket, you should pass the optional ``options`` property:
    {
        "url": "http://testserver.org/oscarapi/products/209/",
        "quantity": 6,
        "options": [{
            "option": "http://testserver.org/oscarapi/options/1/",
            "value": "some value"
        }]
    }
    """

    add_product_serializer_class = AddProductSerializer
    serializer_class = BasketSerializer

    def validate(
        self, basket, product, quantity, options
    ):  # pylint: disable=unused-argument
        availability = basket.strategy.fetch_for_product(product).availability

        # check if product is available at all
        if not availability.is_available_to_buy:
            return False, availability.message

        current_qty = basket.product_quantity(product)
        desired_qty = current_qty + quantity

        # check if we can buy this quantity
        allowed, message = availability.is_purchase_permitted(desired_qty)
        if not allowed:
            return False, message

        # check if there is a limit on amount
        allowed, message = basket.is_quantity_allowed(desired_qty)
        if not allowed:
            return False, message
        return True, None

    def post(self, request, *args, **kwargs):  # pylint: disable=redefined-builtin
        p_ser = self.add_product_serializer_class(
            data=request.data, context={"request": request}
        )
        if p_ser.is_valid():
            basket = operations.get_basket(request)
            product = p_ser.validated_data["url"]
            quantity = p_ser.validated_data["quantity"]
            options = p_ser.validated_data.get("options", [])

            basket_valid, message = self.validate(basket, product, quantity, options)
            if not basket_valid:
                return Response(
                    {"reason": message}, status=status.HTTP_406_NOT_ACCEPTABLE
                )

            basket.add_product(product, quantity=quantity, options=options)

            signals.basket_addition.send(
                sender=self, product=product, user=request.user, request=request
            )

            operations.apply_offers(request, basket)
            ser = self.serializer_class(basket, context={"request": request})
            return Response(ser.data)

        return Response({"reason": p_ser.errors}, status=status.HTTP_406_NOT_ACCEPTABLE)


class AddVoucherView(APIView):
    """
    Add a voucher to the basket.

    POST(vouchercode)
    {
        "vouchercode": "kjadjhgadjgh7667"
    }

    Will return 200 and the voucher as json if successful.
    If unsuccessful, will return 406 with the error.
    """

    add_voucher_serializer_class = VoucherAddSerializer
    serializer_class = VoucherSerializer

    def post(self, request, *args, **kwargs):  # pylint: disable=redefined-builtin
        v_ser = self.add_voucher_serializer_class(
            data=request.data, context={"request": request}
        )
        if v_ser.is_valid():
            basket = operations.get_basket(request)

            voucher = v_ser.instance
            basket.vouchers.add(voucher)

            signals.voucher_addition.send(sender=None, basket=basket, voucher=voucher)

            # Recalculate discounts to see if the voucher gives any
            operations.apply_offers(request, basket)
            discounts_after = basket.offer_applications

            # Look for discounts from this new voucher
            for discount in discounts_after:
                if discount["voucher"] and discount["voucher"] == voucher:
                    break
            else:
                basket.vouchers.remove(voucher)
                return Response(
                    {
                        "reason": _(
                            "Your basket does not qualify for a voucher discount"
                        )
                    },  # noqa
                    status=status.HTTP_406_NOT_ACCEPTABLE,
                )

            ser = self.serializer_class(voucher, context={"request": request})
            return Response(ser.data)

        return Response(v_ser.errors, status=status.HTTP_406_NOT_ACCEPTABLE)


class ShippingMethodView(APIView):
    """
    GET:
    Retrieve shipping methods available to the user, basket, request combination

    POST(shipping_address):
    {
        "country": "http://127.0.0.1:8000/oscarapi/countries/NL/",
        "first_name": "Henk",
        "last_name": "Van den Heuvel",
        "line1": "Roemerlaan 44",
        "line2": "",
        "line3": "",
        "line4": "Kroekingen",
        "notes": "Niet STUK MAKEN OK!!!!",
        "phone_number": "+31 26 370 4887",
        "postcode": "7777KK",
        "state": "Gerendrecht",
        "title": "Mr"
    }

    Post a shipping_address if your shipping methods are dependent on the
    address.
    """

    serializer_class = ShippingAddressSerializer
    shipping_method_serializer_class = ShippingMethodSerializer

    def _get(self, request, shipping_address=None):  # pylint: disable=redefined-builtin
        basket = operations.get_basket(request)
        shiping_methods = Repository().get_shipping_methods(
            basket=basket,
            user=request.user,
            shipping_addr=shipping_address,
            request=request,
        )
        ser = self.shipping_method_serializer_class(
            shiping_methods, many=True, context={"basket": basket}
        )
        return Response(ser.data)

    def get(self, request, *args, **kwargs):  # pylint: disable=redefined-builtin
        """
        Get the available shipping methods and their cost for this order.

        GET:
        A list of shipping method details and the prices.
        """
        return self._get(request)

    def post(self, request, *args, **kwargs):  # pylint: disable=redefined-builtin
        s_ser = self.serializer_class(data=request.data, context={"request": request})
        if s_ser.is_valid():
            shipping_address = ShippingAddress(**s_ser.validated_data)
            return self._get(request, shipping_address=shipping_address)

        return Response(s_ser.errors, status=status.HTTP_406_NOT_ACCEPTABLE)


class LineList(BasketPermissionMixin, generics.ListCreateAPIView):
    """
    Api for adding lines to a basket.

    Permission will be checked,
    Regular users may only access their own basket,
    staff users may access any basket.

    GET:
    A list of basket lines.

    POST(basket, line_reference, product, stockrecord,
         quantity, price_currency, price_excl_tax, price_incl_tax):
    Add a line to the basket, example::

        {
            "basket": "http://127.0.0.1:8000/oscarapi/baskets/100/",
            "line_reference": "234_345",
            "product": "http://127.0.0.1:8000/oscarapi/products/209/",
            "stockrecord":
                "http://127.0.0.1:8000/ooscarapi/products/209/stockercords/1/",
            "quantity": 3,
            "price_currency": "EUR",
            "price_excl_tax": "100.0",
            "price_incl_tax": "121.0"
        }
    """

    permission_classes = (permissions.RequestAllowsAccessTo,)
    serializer_class = BasketLineSerializer
    queryset = Line.objects.all()

    def get_queryset(self):
        basket_pk = self.kwargs.get("pk")
        basket = self.check_basket_permission(self.request, basket_pk=basket_pk)
        prepped_basket = operations.assign_basket_strategy(basket, self.request)
        return prepped_basket.all_lines()

    def post(
        self, request, pk, format=None, *args, **kwargs
    ):  # pylint: disable=redefined-builtin,arguments-differ
        data_basket = self.get_data_basket(request.data, format)
        self.check_basket_permission(request, basket=data_basket)

        url_basket = self.check_basket_permission(request, basket_pk=pk)

        if url_basket != data_basket:
            raise exceptions.NotAcceptable(
                _("Target basket inconsistent %s != %s")
                % (url_basket.pk, data_basket.pk)
            )
        return super(LineList, self).post(request, format=format)


class BasketLineDetail(generics.RetrieveUpdateDestroyAPIView):
    """
    Only the field `quantity` can be changed in this view.
    All other fields are readonly.
    """

    queryset = Line.objects.all()
    serializer_class = BasketLineSerializer
    permission_classes = (permissions.RequestAllowsAccessTo,)

    def get_queryset(self):
        basket_pk = self.kwargs.get("basket_pk")
        basket = generics.get_object_or_404(operations.editable_baskets(), pk=basket_pk)
        prepped_basket = operations.prepare_basket(basket, self.request)
        if operations.request_allows_access_to_basket(self.request, prepped_basket):
            return prepped_basket.all_lines()
        else:
            return self.queryset.none()



from django.conf import settings
from django.contrib.auth import get_user_model

from rest_framework import generics, status
from rest_framework.exceptions import MethodNotAllowed
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from oscar.apps.customer.signals import user_registered
from oscar.core.loading import get_class, get_model

from oscarapi.utils.session import login_and_upgrade_session
from oscarapi.utils.loading import get_api_classes
from oscarapi.basket import operations


LoginSerializer, UserSerializer, RegisterUserSerializer = get_api_classes(
    "serializers.login", ["LoginSerializer", "UserSerializer", "RegisterUserSerializer"]
)
RegisterUserMixin = get_class("customer.mixins", "RegisterUserMixin")

Basket = get_model("basket", "Basket")
User = get_user_model()

__all__ = ("LoginView", "UserDetail")


class LoginView(APIView):
    """
    Api for logging in users.

    DELETE:
    Log the user out by destroying the session.
    Anonymous users will have their cart destroyed as well, because there is
    no way to reach it anymore

    POST(username, password):
    1. The user will be authenticated. The next steps will only be
       performed is login is successful. Logging in logged in users results in
       405.
    2. The anonymous cart will be merged with the private cart associated with
       that authenticated user.
    3. A new session will be started, this session identifies the authenticated
       user for the duration of the session, without further need for
       authentication.
    4. The new, merged cart will be associated with this session.
    5. The anonymous session will be terminated.
    6. A response will be issued containing the new session id as a header
       (only when the request contained the session header as well).

    GET:
    Get the details of the logged in user. Can be enabled/disabled with the
    OSCARAPI_EXPOSE_USER_DETAILS setting. If more details are needed,
    use the ``OSCARAPI_USER_FIELDS`` setting to change the fields the
    ``UserSerializer`` will render.
    """

    serializer_class = LoginSerializer

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            if getattr(settings, "OSCARAPI_EXPOSE_USER_DETAILS", False):
                ser = UserSerializer(request.user, many=False)
                return Response(ser.data)
            return Response(status=status.HTTP_204_NO_CONTENT)
        raise MethodNotAllowed("GET")

    def merge_baskets(self, anonymous_basket, basket):
        "Hook to enforce rules when merging baskets."
        basket.merge(anonymous_basket)

    def post(self, request, *args, **kwargs):
        ser = self.serializer_class(data=request.data)
        if ser.is_valid():

            anonymous_basket = operations.get_anonymous_basket(request)

            user = ser.instance

            # refuse to login logged in users, to avoid attaching sessions to
            # multiple users at the same time.
            if request.user.is_authenticated:
                return Response(
                    {"detail": "Session is in use, log out first"},
                    status=status.HTTP_405_METHOD_NOT_ALLOWED,
                )

            request.user = user

            login_and_upgrade_session(request._request, user)

            # merge anonymous basket with authenticated basket.
            basket = operations.get_user_basket(user)
            if anonymous_basket is not None:
                self.merge_baskets(anonymous_basket, basket)

            operations.store_basket_in_session(basket, request.session)

            return Response("")

        return Response(ser.errors, status=status.HTTP_401_UNAUTHORIZED)

    def delete(self, request, *args, **kwargs):
        """
        Destroy the session.

        for anonymous users that means having their basket destroyed as well,
        because there is no way to reach it otherwise.
        """
        request = request._request
        if request.user.is_anonymous:
            basket = operations.get_anonymous_basket(request)
            if basket:
                operations.flush_and_delete_basket(basket)

        request.session.clear()
        request.session.delete()
        request.session = None

        return Response("")


class UserDetail(generics.RetrieveAPIView):
    serializer_class = UserSerializer
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        return User.objects.filter(pk=self.request.user.pk)

    def get(self, request, *args, **kwargs):
        if getattr(settings, "OSCARAPI_EXPOSE_USER_DETAILS", False):
            return super().get(request, *args, **kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RegistrationView(APIView, RegisterUserMixin):
    """
    API for registering users

    POST(email, password1, password2):
    {
        "email": "user@my-domain.com",
        "password1": "MyVerySecretPassword123"
        "password2": "MyVerySecretPassword123"
    }

    Will create a new user when the user with the specific email does
    not exist (HTTP_201_CREATED). It will also send a user_registered signal.

    It won't login the newly created user, You can do this with the login API.
    """

    serializer_class = RegisterUserSerializer

    def post(self, request, *args, **kwargs):
        if not getattr(settings, "OSCARAPI_ENABLE_REGISTRATION", False):
            return Response("Unauthorized", status=status.HTTP_401_UNAUTHORIZED)

        ser = self.serializer_class(data=request.data)

        if ser.is_valid():
            # create the user
            user = ser.save()

            if getattr(settings, "OSCAR_SEND_REGISTRATION_EMAIL", False):
                self.send_registration_email(user)
            # send the same signal as oscar is sending
            user_registered.send(sender=self, request=request, user=user)

            return Response(user.email, status=status.HTTP_201_CREATED)

        return Response(ser.errors, status=status.HTTP_400_BAD_REQUEST)



from importlib import reload as reload_module

import json
from re import match

from django.contrib.auth import get_user_model
from django.urls import reverse, NoReverseMatch
from django.http import SimpleCookie
from django.test import TestCase


User = get_user_model()


class APITest(TestCase):
    longMessage = True

    def setUp(self):
        user = User.objects.create_user(
            username="admin", email="admin@admin.admin", password="admin"
        )
        user.is_staff = True
        user.is_superuser = True
        user.save()

        user = User.objects.create_user(
            username="nobody", email="nobody@nobody.niks", password="nobody"
        )
        user.is_staff = False
        user.is_superuser = False
        user.save()

        user = User.objects.create_user(
            username="somebody", email="somebody@nobody.niks", password="somebody"
        )
        user.is_staff = False
        user.is_superuser = False
        user.save()

    def login(self, username, password):
        result = self.client.login(username=username, password=password)
        self.assertTrue(result, "%s should be able to log in" % username)
        return True

    def hlogin(self, username, password, session_id):
        response = self.post(
            "api-login", session_id, username=username, password=password
        )
        self.assertEqual(
            response.status_code,
            200,
            "%s should be able to login via the api" % username,
        )
        return True

    def api_call(self, url_name, method, session_id=None, authenticated=False, **data):
        try:
            url = reverse(url_name)
        except NoReverseMatch:
            url = url_name
        method = getattr(self.client, method.lower())
        kwargs = {"content_type": "application/json"}
        if session_id is not None:
            auth_type = "AUTH" if authenticated else "ANON"
            kwargs["HTTP_SESSION_ID"] = "SID:%s:testserver:%s" % (auth_type, session_id)

        response = None
        if data:
            response = method(url, json.dumps(data), **kwargs)
        else:
            response = method(url, **kwargs)
        # throw away cookies when using session_id authentication
        if session_id is not None:
            self.client.cookies = SimpleCookie()

        return response

    def get(self, url_name, session_id=None, authenticated=False):
        return self.api_call(
            url_name, "GET", session_id=session_id, authenticated=authenticated
        )

    def post(self, url_name, session_id=None, authenticated=False, **data):
        return self.api_call(
            url_name, "POST", session_id=session_id, authenticated=authenticated, **data
        )

    def put(self, url_name, session_id=None, authenticated=False, **data):
        return self.api_call(
            url_name, "PUT", session_id=session_id, authenticated=authenticated, **data
        )

    def patch(self, url_name, session_id=None, authenticated=False, **data):
        return self.api_call(
            url_name,
            "PATCH",
            session_id=session_id,
            authenticated=authenticated,
            **data
        )

    def delete(self, url_name, session_id=None, authenticated=False):
        return self.api_call(
            url_name, "DELETE", session_id=session_id, authenticated=authenticated
        )

    def tearDown(self):
        User.objects.get(username="admin").delete()
        User.objects.get(username="nobody").delete()

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        self._response = ParsedResponse(response, self)

    @staticmethod
    def reload_modules(modules=()):
        for module in modules:
            reload_module(module)


class ParsedResponse(object):
    def __init__(self, response, unittestcase):
        self.response = response
        self.t = unittestcase

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        self._response = response
        self.status_code = response.status_code
        try:
            self.body = response.data
        except Exception:
            self.body = None

    def __getattr__(self, name):
        return getattr(self._response, name)

    def __getitem__(self, name):
        return self.body[name]

    def __len__(self):
        return len(self.body)

    def assertStatusEqual(self, code, message=None):
        self.t.assertEqual(self.status_code, code, message)

    def assertValueEqual(self, value_name, value, message=None):
        self.t.assertEqual(self[value_name], value, message)

    def assertObjectIdEqual(self, value_name, value, message=None):
        pattern = r".*?%s.*?/(?P<object_id>\d+)/?" % reverse("api-root")
        m = match(pattern, self[value_name])
        if m:
            object_id = int(m.groupdict()["object_id"])
        else:
            object_id = None
        self.t.assertEqual(object_id, value, message)

    def __str__(self):
        return str(self._response)



from rest_framework.exceptions import _get_error_details


class FieldError(Exception):
    default_code = "field_error"

    def __init__(self, detail, code=None):
        if code is None:
            code = self.default_code

        self.code = code
        self.detail = _get_error_details(detail, code)
        super(FieldError, self).__init__(self.detail)



from django.conf import settings
from oscar.core.loading import _pluck_classes, _import_module

OSCARAPI_OVERRIDE_MODULES = getattr(settings, "OSCARAPI_OVERRIDE_MODULES", [])


def oscarapi_class_loader(module_label, classnames, module_prefix="oscarapi"):
    """Oscarapi uses a bit simpler method of overrides"""
    default_module_name = "%s.%s" % (module_prefix, module_label)
    default_module = _import_module(default_module_name, classnames)
    class_search_modules = []

    # load all modules to search for classes in
    for module_name in OSCARAPI_OVERRIDE_MODULES:  # could be empty
        override_module_name = "%s.%s" % (module_name, module_label)
        override_module = _import_module(override_module_name, classnames)
        if override_module is not None:
            class_search_modules.append(override_module)

    # add the module from the standard oscarapi package
    class_search_modules.append(default_module)
    return _pluck_classes(class_search_modules, classnames)


def get_api_classes(module_label, classnames):
    return oscarapi_class_loader(module_label, classnames, "oscarapi")


def get_api_class(module_label, classname):
    return get_api_classes(module_label, [classname])[0]



import hashlib


def file_hash(content):
    """
    file_hash will hash the content of a file.

    >>> from io import BytesIO
    >>> long_text = "Harrie en Henk (Hent)" * 879
    >>> file_hash(BytesIO(long_text.encode()))
    'cf3c7be4778cded29fab4b66622c5e8af91d01a9'
    """
    sha1 = getattr(content, "sha1", None)
    if sha1 is not None:
        return sha1

    hasher = hashlib.sha1()
    buf = content.read(65536)
    while len(buf) > 0:
        hasher.update(buf)
        buf = content.read(65536)

    content.seek(0)
    return hasher.hexdigest()



default_app_config = "oscarapi.apps.OscarAPIConfig"

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata as importlib_metadata

version = importlib_metadata.version("django-oscar-api")



def entity_internal_value(attribute, value):
    raise NotImplementedError(
        "Writable Entity support requires a manual implementation, "
        "because it is not possible to guess how the value "
        "sent should be interpreted. You can override "
        "'oscarapi.serializers.hooks.entity_internal_value' to provide an implementation"
    )



from oscar.core.loading import get_model

from rest_framework import generics, response, status, views

from oscarapi.basket.operations import request_allows_access_to_basket
from oscarapi.permissions import IsOwner
from oscarapi.utils.loading import get_api_classes
from oscarapi.signals import oscarapi_post_checkout
from oscarapi.views.utils import parse_basket_from_hyperlink

Order = get_model("order", "Order")
OrderLine = get_model("order", "Line")
OrderLineAttribute = get_model("order", "LineAttribute")
UserAddress = get_model("address", "UserAddress")
(
    CheckoutSerializer,
    OrderLineAttributeSerializer,
    OrderLineSerializer,
    OrderSerializer,
    UserAddressSerializer,
) = get_api_classes(
    "serializers.checkout",
    [
        "CheckoutSerializer",
        "OrderLineAttributeSerializer",
        "OrderLineSerializer",
        "OrderSerializer",
        "UserAddressSerializer",
    ],
)

__all__ = (
    "CheckoutView",
    "OrderList",
    "OrderDetail",
    "OrderLineList",
    "OrderLineDetail",
    "OrderLineAttributeDetail",
    "UserAddressList",
    "UserAddressDetail",
)


class OrderList(generics.ListAPIView):
    serializer_class = OrderSerializer
    permission_classes = (IsOwner,)

    def get_queryset(self):
        qs = Order.objects.all()
        return qs.filter(user=self.request.user)


class OrderDetail(generics.RetrieveAPIView):
    queryset = Order.objects.all()
    serializer_class = OrderSerializer
    permission_classes = (IsOwner,)


class OrderLineList(generics.ListAPIView):
    queryset = OrderLine.objects.all()
    serializer_class = OrderLineSerializer

    def get_queryset(self):
        pk = self.kwargs.get("pk")
        user = self.request.user
        return super().get_queryset().filter(order_id=pk, order__user=user)


class OrderLineDetail(generics.RetrieveAPIView):
    queryset = OrderLine.objects.all()
    serializer_class = OrderLineSerializer

    def get_queryset(self):
        return super().get_queryset().filter(order__user=self.request.user)


class OrderLineAttributeDetail(generics.RetrieveAPIView):
    queryset = OrderLineAttribute.objects.all()
    serializer_class = OrderLineAttributeSerializer


class CheckoutView(views.APIView):
    """
    Prepare an order for checkout.

    POST(basket, shipping_address,
         [total, shipping_method_code, shipping_charge, billing_address]):
    {
        "basket": "http://testserver/oscarapi/baskets/1/",
        "guest_email": "foo@example.com",
        "total": "100.0",
        "shipping_charge": {
            "currency": "EUR",
            "excl_tax": "10.0",
            "tax": "0.6"
        },
        "shipping_method_code": "no-shipping-required",
        "shipping_address": {
            "country": "http://127.0.0.1:8000/oscarapi/countries/NL/",
            "first_name": "Henk",
            "last_name": "Van den Heuvel",
            "line1": "Roemerlaan 44",
            "line2": "",
            "line3": "",
            "line4": "Kroekingen",
            "notes": "Niet STUK MAKEN OK!!!!",
            "phone_number": "+31 26 370 4887",
            "postcode": "7777KK",
            "state": "Gerendrecht",
            "title": "Mr"
        }
        "billing_address": {
            "country": country_url,
            "first_name": "Jos",
            "last_name": "Henken",
            "line1": "Boerderijstraat 19",
            "line2": "",
            "line3": "",
            "line4": "Zwammerdam",
            "notes": "",
            "phone_number": "+31 27 112 9800",
            "postcode": "6666LL",
            "state": "Gerendrecht",
            "title": "Mr"
         }
    }
    returns the order object.
    """

    order_serializer_class = OrderSerializer
    serializer_class = CheckoutSerializer

    def post(self, request, format=None, *args, **kwargs):
        # TODO: Make it possible to create orders with options.
        # at the moment, no options are passed to this method, which means they
        # are also not created.

        basket = parse_basket_from_hyperlink(request.data, format)

        if not request_allows_access_to_basket(request, basket):
            return response.Response(
                "Unauthorized", status=status.HTTP_401_UNAUTHORIZED
            )

        c_ser = self.serializer_class(data=request.data, context={"request": request})

        if c_ser.is_valid():
            order = c_ser.save()
            basket.freeze()
            o_ser = self.order_serializer_class(order, context={"request": request})

            resp = response.Response(o_ser.data)

            oscarapi_post_checkout.send(
                sender=self,
                order=order,
                user=request.user,
                request=request,
                response=resp,
            )
            return resp

        return response.Response(c_ser.errors, status.HTTP_406_NOT_ACCEPTABLE)


class UserAddressList(generics.ListCreateAPIView):
    serializer_class = UserAddressSerializer
    permission_classes = (IsOwner,)

    def get_queryset(self):
        return UserAddress.objects.filter(user=self.request.user)


class UserAddressDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = UserAddressSerializer
    permission_classes = (IsOwner,)

    def get_queryset(self):
        return UserAddress.objects.filter(user=self.request.user)



from setuptools import find_packages, setup

__version__ = "3.1.1"

setup(
    # package name in pypi
    name="django-oscar-api",
    # extract version from module.
    version=__version__,
    description="REST API module for django-oscar",
    long_description=open("README.rst").read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    python_requires=">=3.6",
    keywords="",
    author="Lars van de Kerkhof, Martijn Jacobs",
    author_email="lars@permanentmarkers.nl, martijn@devopsconsulting.nl",
    url="https://github.com/django-oscar/django-oscar-api",
    license="BSD",
    packages=find_packages(
        exclude=[
            "*tests.unit",
            "*tests.serializers*",
            "*tests.doctests*",
            "*fixtures",
            "*fixtures*",
            "*sandbox*",
        ]
    ),
    # include non python files
    include_package_data=True,
    zip_safe=False,
    # specify dependencies
    install_requires=[
        "setuptools",
        "django-oscar>=3.0",
        "Django>=2.2.24",  # CVE-2021-33203
        "djangorestframework>=3.9",  # first version to support Django 2.2
    ],
    # mark test target to require extras.
    extras_require={
        "dev": ["coverage", "mock", "twine", "wheel", "easy_thumbnails"],
        "lint": ["flake8", "flake8-black", "flake8-bugbear", "black>=19.10b0"],
    },
)



import functools


from oscar.core.loading import get_class, get_model

from rest_framework import generics

from oscarapi import permissions
from oscarapi.basket.operations import (
    assign_basket_strategy,
    editable_baskets,
    get_anonymous_basket,
    prepare_basket,
)
from oscarapi.utils.loading import get_api_classes, get_api_class

from .utils import QuerySetList

__all__ = (
    "BasketList",
    "BasketDetail",
    "LineAttributeDetail",
    "OptionList",
    "OptionDetail",
    "CountryList",
    "CountryDetail",
)

Basket = get_model("basket", "Basket")
LineAttribute = get_model("basket", "LineAttribute")
Product = get_model("catalogue", "Product")
Option = get_model("catalogue", "Option")
Country = get_model("address", "Country")
Range = get_model("offer", "Range")

Selector = get_class("partner.strategy", "Selector")
CountrySerializer = get_api_class("serializers.checkout", "CountrySerializer")
(
    BasketSerializer,
    LineAttributeSerializer,
) = get_api_classes(  # pylint: disable=unbalanced-tuple-unpacking
    "serializers.basket", ["BasketSerializer", "LineAttributeSerializer"]
)
(
    RangeSerializer,
    OptionSerializer,
) = get_api_classes(  # pylint: disable=unbalanced-tuple-unpacking
    "serializers.product", ["RangeSerializer", "OptionSerializer"]
)


class CountryList(generics.ListAPIView):
    serializer_class = CountrySerializer
    queryset = Country.objects.all()


class CountryDetail(generics.RetrieveAPIView):
    serializer_class = CountrySerializer
    queryset = Country.objects.all()


class BasketList(generics.ListAPIView):
    """
    Retrieve all baskets that belong to the current (authenticated) user.
    """

    serializer_class = BasketSerializer
    queryset = editable_baskets()

    def get_queryset(self):
        qs = super(BasketList, self).get_queryset()

        if self.request.user.is_authenticated:
            qs = qs.filter(owner=self.request.user)
            mapped_with_baskets = list(
                map(functools.partial(assign_basket_strategy, request=self.request), qs)
            )
        else:  # anonymous users have max 1 basket.
            basket = get_anonymous_basket(self.request)
            mapped_with_baskets = [prepare_basket(basket, self.request)]

        return QuerySetList(mapped_with_baskets, qs)


class BasketDetail(generics.RetrieveUpdateDestroyAPIView):
    serializer_class = BasketSerializer
    permission_classes = (permissions.RequestAllowsAccessTo,)
    queryset = editable_baskets()

    def get_object(self):
        basket = super(BasketDetail, self).get_object()
        return assign_basket_strategy(basket, self.request)


class LineAttributeDetail(generics.RetrieveUpdateAPIView):
    queryset = LineAttribute.objects.all()
    serializer_class = LineAttributeSerializer
    permission_classes = (permissions.RequestAllowsAccessTo,)


class OptionList(generics.ListAPIView):
    queryset = Option.objects.all()
    serializer_class = OptionSerializer


class OptionDetail(generics.RetrieveAPIView):
    queryset = Option.objects.all()
    serializer_class = OptionSerializer


class RangeList(generics.ListAPIView):
    queryset = Range.objects.all()
    serializer_class = RangeSerializer


class RangeDetail(generics.RetrieveAPIView):
    queryset = Range.objects.all()
    serializer_class = RangeSerializer



# pylint: disable=unbalanced-tuple-unpacking
from rest_framework import generics
from rest_framework.response import Response

from oscar.core.loading import get_class, get_model

from oscarapi.utils.categories import find_from_full_slug
from oscarapi.utils.loading import get_api_classes, get_api_class

Selector = get_class("partner.strategy", "Selector")
(
    CategorySerializer,
    ProductLinkSerializer,
    ProductSerializer,
    ProductStockRecordSerializer,
    AvailabilitySerializer,
) = get_api_classes(
    "serializers.product",
    [
        "CategorySerializer",
        "ProductLinkSerializer",
        "ProductSerializer",
        "ProductStockRecordSerializer",
        "AvailabilitySerializer",
    ],
)

PriceSerializer = get_api_class("serializers.checkout", "PriceSerializer")


__all__ = ("ProductList", "ProductDetail", "ProductPrice", "ProductAvailability")

Product = get_model("catalogue", "Product")
Category = get_model("catalogue", "Category")
StockRecord = get_model("partner", "StockRecord")


class ProductList(generics.ListAPIView):
    queryset = Product.objects.all()
    serializer_class = ProductLinkSerializer

    def get_queryset(self):
        """
        Allow filtering on structure so standalone and parent products can
        be selected separately, eg::

            http://127.0.0.1:8000/api/products/?structure=standalone

        or::

            http://127.0.0.1:8000/api/products/?structure=parent
        """
        qs = super(ProductList, self).get_queryset()
        structure = self.request.query_params.get("structure")
        if structure is not None:
            return qs.filter(structure=structure)

        return qs


class ProductDetail(generics.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = ProductSerializer


class ProductPrice(generics.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = PriceSerializer

    def get(
        self, request, pk=None, *args, **kwargs
    ):  # pylint: disable=redefined-builtin,arguments-differ
        product = self.get_object()
        strategy = Selector().strategy(request=request, user=request.user)
        ser = PriceSerializer(
            strategy.fetch_for_product(product).price, context={"request": request}
        )
        return Response(ser.data)


class ProductStockRecords(generics.ListAPIView):
    serializer_class = ProductStockRecordSerializer
    queryset = StockRecord.objects.all()

    def get_queryset(self):
        product_pk = self.kwargs.get("pk")
        return super().get_queryset().filter(product_id=product_pk)


class ProductStockRecordDetail(generics.RetrieveAPIView):
    serializer_class = ProductStockRecordSerializer
    queryset = StockRecord.objects.all()


class ProductAvailability(generics.RetrieveAPIView):
    queryset = Product.objects.all()
    serializer_class = AvailabilitySerializer

    def get(
        self, request, pk=None, *args, **kwargs
    ):  # pylint: disable=redefined-builtin,arguments-differ
        product = self.get_object()
        strategy = Selector().strategy(request=request, user=request.user)
        ser = AvailabilitySerializer(
            strategy.fetch_for_product(product).availability,
            context={"request": request},
        )
        return Response(ser.data)


class CategoryList(generics.ListAPIView):
    queryset = Category.get_root_nodes()
    serializer_class = CategorySerializer

    def get_queryset(self):
        breadcrumb_path = self.kwargs.get("breadcrumbs", None)
        if breadcrumb_path is None:
            return super(CategoryList, self).get_queryset()

        return find_from_full_slug(breadcrumb_path, separator="/").get_children()


class CategoryDetail(generics.RetrieveAPIView):
    queryset = Category.objects.all()
    serializer_class = CategorySerializer



from django.core.exceptions import ValidationError

from django.contrib.auth import get_user_model, authenticate, password_validation
from rest_framework import serializers

from oscarapi.utils.settings import overridable


User = get_user_model()


def field_length(fieldname):
    field = next(field for field in User._meta.fields if field.name == fieldname)
    return field.max_length


class UserSerializer(serializers.ModelSerializer):
    date_joined = serializers.DateTimeField(read_only=True)

    class Meta:
        model = User
        fields = overridable(
            "OSCARAPI_USER_FIELDS",
            default=(User.USERNAME_FIELD, "email", "date_joined"),
        )


class LoginSerializer(serializers.Serializer):
    username = serializers.CharField(
        max_length=field_length(User.USERNAME_FIELD), required=True
    )
    password = serializers.CharField(
        max_length=field_length("password"),
        required=True,
        style={"input_type": "password"},
    )

    def validate(self, attrs):
        user = authenticate(username=attrs["username"], password=attrs["password"])
        if user is None:
            raise serializers.ValidationError("invalid login")
        elif not user.is_active:
            raise serializers.ValidationError("Can not log in as inactive user")

        # set instance to the user so we can use this in the view
        self.instance = user
        return attrs


class RegisterUserSerializer(serializers.Serializer):
    email = serializers.EmailField(max_length=field_length("email"), required=True)
    password1 = serializers.CharField(
        max_length=field_length("password"),
        required=True,
        style={"input_type": "password"},
    )
    password2 = serializers.CharField(
        max_length=field_length("password"),
        required=True,
        style={"input_type": "password"},
    )

    def validate(self, attrs):
        if User.objects.filter(email=attrs["email"]).exists():
            raise serializers.ValidationError("A user with this email already exists")

        if attrs["password1"] != attrs["password2"]:
            raise serializers.ValidationError("Passwords do not match")

        try:
            password_validation.validate_password(attrs["password1"])
        except ValidationError as e:
            raise serializers.ValidationError(str(e))

        return attrs

    def create_user(self, email, password):
        # this is a separate method so it's easy to override
        return User.objects.create_user(username=email, email=email, password=password)

    def save(self):
        email = self.validated_data["email"]
        password = self.validated_data["password1"]
        return self.create_user(email, password)



from django.urls import reverse


def get_domain(request):
    """
    Get the domain name parsed from a hostname:port string

    >>> class FakeRequest(object):
    ...     def __init__(self, url):
    ...         self.url = url
    ...     def get_host(self):
    ...         return self.url
    >>> req = FakeRequest("example.com:5984")
    >>> get_domain(req)
    'example.com'
    """
    return request.get_host().split(":")[0]


class IsApiRequest(object):
    @staticmethod
    def is_api_request(request):
        path = request.path.lower()
        api_root = reverse("api-root").lower()
        return path.startswith(api_root)



# pylint: disable=unbalanced-tuple-unpacking
from django.conf import settings
from django.urls import include, path, re_path
from rest_framework.urlpatterns import format_suffix_patterns

from oscarapi.utils.loading import get_api_classes, get_api_class

api_root = get_api_class("views.root", "api_root")
(LoginView, UserDetail, RegistrationView) = get_api_classes(
    "views.login", ["LoginView", "UserDetail", "RegistrationView"]
)
(
    BasketView,
    AddProductView,
    AddVoucherView,
    ShippingMethodView,
    LineList,
    BasketLineDetail,
) = get_api_classes(
    "views.basket",
    [
        "BasketView",
        "AddProductView",
        "AddVoucherView",
        "ShippingMethodView",
        "LineList",
        "BasketLineDetail",
    ],
)

(StockRecordDetail, PartnerList, PartnerDetail) = get_api_classes(
    "views.admin.partner", ["StockRecordDetail", "PartnerList", "PartnerDetail"]
)

(
    BasketList,
    BasketDetail,
    LineAttributeDetail,
    OptionList,
    OptionDetail,
    CountryList,
    CountryDetail,
    RangeList,
    RangeDetail,
) = get_api_classes(
    "views.basic",
    [
        "BasketList",
        "BasketDetail",
        "LineAttributeDetail",
        "OptionList",
        "OptionDetail",
        "CountryList",
        "CountryDetail",
        "RangeList",
        "RangeDetail",
    ],
)

(
    ProductList,
    ProductDetail,
    ProductStockRecords,
    ProductStockRecordDetail,
    ProductPrice,
    ProductAvailability,
    CategoryList,
    CategoryDetail,
) = get_api_classes(
    "views.product",
    [
        "ProductList",
        "ProductDetail",
        "ProductStockRecords",
        "ProductStockRecordDetail",
        "ProductPrice",
        "ProductAvailability",
        "CategoryList",
        "CategoryDetail",
    ],
)

(
    CheckoutView,
    OrderList,
    OrderDetail,
    OrderLineList,
    OrderLineDetail,
    OrderLineAttributeDetail,
    UserAddressList,
    UserAddressDetail,
) = get_api_classes(
    "views.checkout",
    [
        "CheckoutView",
        "OrderList",
        "OrderDetail",
        "OrderLineList",
        "OrderLineDetail",
        "OrderLineAttributeDetail",
        "UserAddressList",
        "UserAddressDetail",
    ],
)

(
    ProductAdminList,
    ProductAdminDetail,
    ProductClassAdminList,
    ProductClassAdminDetail,
    ProductAttributeAdminList,
    ProductAttributeAdminDetail,
    AttributeOptionGroupAdminList,
    AttributeOptionGroupAdminDetail,
    CategoryAdminList,
    CategoryAdminDetail,
) = get_api_classes(
    "views.admin.product",
    [
        "ProductAdminList",
        "ProductAdminDetail",
        "ProductClassAdminList",
        "ProductClassAdminDetail",
        "ProductAttributeAdminList",
        "ProductAttributeAdminDetail",
        "AttributeOptionGroupAdminList",
        "AttributeOptionGroupAdminDetail",
        "CategoryAdminList",
        "CategoryAdminDetail",
    ],
)

(
    OrderAdminList,
    OrderAdminDetail,
    OrderLineAdminList,
    OrderLineAdminDetail,
    OrderLineAttributeAdminDetail,
) = get_api_classes(
    "views.admin.order",
    [
        "OrderAdminList",
        "OrderAdminDetail",
        "OrderLineAdminList",
        "OrderLineAdminDetail",
        "OrderLineAttributeAdminDetail",
    ],
)

(UserAdminList, UserAdminDetail) = get_api_classes(
    "views.admin.user", ["UserAdminList", "UserAdminDetail"]
)

urlpatterns = [
    path("", api_root, name="api-root"),
    path("register/", RegistrationView.as_view(), name="api-register"),
    path("login/", LoginView.as_view(), name="api-login"),
    path("basket/", BasketView.as_view(), name="api-basket"),
    path(
        "basket/add-product/", AddProductView.as_view(), name="api-basket-add-product"
    ),
    path(
        "basket/add-voucher/", AddVoucherView.as_view(), name="api-basket-add-voucher"
    ),
    path(
        "basket/shipping-methods/",
        ShippingMethodView.as_view(),
        name="api-basket-shipping-methods",
    ),
    path("baskets/", BasketList.as_view(), name="basket-list"),
    path("baskets/<int:pk>/", BasketDetail.as_view(), name="basket-detail"),
    path("baskets/<int:pk>/lines/", LineList.as_view(), name="basket-lines-list"),
    path(
        "baskets/<int:basket_pk>/lines/<int:pk>/",
        BasketLineDetail.as_view(),
        name="basket-line-detail",
    ),
    path(
        "baskets/<int:basket_pk>/lines/<int:line_pk>/lineattributes/<int:pk>/",
        LineAttributeDetail.as_view(),
        name="lineattribute-detail",
    ),
    path("products/", ProductList.as_view(), name="product-list"),
    path("products/<int:pk>/", ProductDetail.as_view(), name="product-detail"),
    path("products/<int:pk>/price/", ProductPrice.as_view(), name="product-price"),
    path(
        "products/<int:pk>/availability/",
        ProductAvailability.as_view(),
        name="product-availability",
    ),
    path(
        "products/<int:pk>/stockrecords/",
        ProductStockRecords.as_view(),
        name="product-stockrecords",
    ),
    path(
        "products/<int:product_pk>/stockrecords/<int:pk>/",
        ProductStockRecordDetail.as_view(),
        name="product-stockrecord-detail",
    ),
    path("options/", OptionList.as_view(), name="option-list"),
    path("options/<int:pk>/", OptionDetail.as_view(), name="option-detail"),
    path("ranges/", RangeList.as_view(), name="range-list"),
    path("ranges/<int:pk>/", RangeDetail.as_view(), name="range-detail"),
    path("categories/", CategoryList.as_view(), name="category-list"),
    path("categories/<int:pk>/", CategoryDetail.as_view(), name="category-detail"),
    re_path(
        "^categories/(?P<breadcrumbs>.*)/$",
        CategoryList.as_view(),
        name="category-child-list",
    ),
    path("users/<int:pk>/", UserDetail.as_view(), name="user-detail"),
    path("checkout/", CheckoutView.as_view(), name="api-checkout"),
    path("orders/", OrderList.as_view(), name="order-list"),
    path("orders/<int:pk>/", OrderDetail.as_view(), name="order-detail"),
    path("orders/<int:pk>/lines/", OrderLineList.as_view(), name="order-lines-list"),
    path("orderlines/<int:pk>/", OrderLineDetail.as_view(), name="order-lines-detail"),
    path(
        "orderlineattributes/<int:pk>/",
        OrderLineAttributeDetail.as_view(),
        name="order-lineattributes-detail",
    ),
    path("countries/", CountryList.as_view(), name="country-list"),
    re_path(
        r"^countries/(?P<pk>[A-z]{2})/$", CountryDetail.as_view(), name="country-detail"
    ),
    path("useraddresses/", UserAddressList.as_view(), name="useraddress-list"),
    path(
        "useraddresses/<int:pk>/",
        UserAddressDetail.as_view(),
        name="useraddress-detail",
    ),
]

admin_urlpatterns = [
    path("products/", ProductAdminList.as_view(), name="admin-product-list"),
    path(
        "products/<int:pk>/",
        ProductAdminDetail.as_view(),
        name="admin-product-detail",
    ),
    path(
        "productclasses/",
        ProductClassAdminList.as_view(),
        name="admin-productclass-list",
    ),
    path(
        "productclasses/<slug:slug>/",
        ProductClassAdminDetail.as_view(),
        name="admin-productclass-detail",
    ),
    path("categories/", CategoryAdminList.as_view(), name="admin-category-list"),
    path(
        "categories/<int:pk>/",
        CategoryAdminDetail.as_view(),
        name="admin-category-detail",
    ),
    re_path(
        r"^categories/(?P<breadcrumbs>.*)/$",
        CategoryAdminList.as_view(),
        name="admin-category-child-list",
    ),
    path(
        "productattributes/",
        ProductAttributeAdminList.as_view(),
        name="admin-productattribute-list",
    ),
    path(
        "stockrecords/<int:pk>/",
        StockRecordDetail.as_view(),
        name="admin-stockrecord-detail",
    ),
    path("partners/", PartnerList.as_view(), name="admin-partner-list"),
    path("partners/<int:pk>/", PartnerDetail.as_view(), name="partner-detail"),
    path(
        "productattributes/<int:pk>/",
        ProductAttributeAdminDetail.as_view(),
        name="admin-productattribute-detail",
    ),
    path(
        "attributeoptiongroups/",
        AttributeOptionGroupAdminList.as_view(),
        name="admin-attributeoptiongroup-list",
    ),
    path(
        "attributeoptiongroups/<int:pk>/",
        AttributeOptionGroupAdminDetail.as_view(),
        name="admin-attributeoptiongroup-detail",
    ),
    path("orders/", OrderAdminList.as_view(), name="admin-order-list"),
    path("orders/<int:pk>/", OrderAdminDetail.as_view(), name="admin-order-detail"),
    path(
        "orders/<int:pk>/lines/",
        OrderLineAdminList.as_view(),
        name="admin-order-lines-list",
    ),
    path(
        "orderlines/<int:pk>/",
        OrderLineAdminDetail.as_view(),
        name="admin-order-lines-detail",
    ),
    path(
        "orderlineattributes/<int:pk>/",
        OrderLineAttributeAdminDetail.as_view(),
        name="admin-order-lineattributes-detail",
    ),
    path("users/", UserAdminList.as_view(), name="admin-user-list"),
    path("users/<int:pk>/", UserAdminDetail.as_view(), name="admin-user-detail"),
]

if not getattr(settings, "OSCARAPI_BLOCK_ADMIN_API_ACCESS", True):
    urlpatterns.append(path("admin/", include(admin_urlpatterns)))

urlpatterns = format_suffix_patterns(urlpatterns)



import logging
from decimal import Decimal

from django.utils.translation import ugettext as _
from django.contrib.auth import get_user_model
from rest_framework import serializers

from oscar.core.loading import get_model

from oscarapi.basket import operations
from oscarapi.utils.settings import overridable
from oscarapi.serializers.fields import (
    DrillDownHyperlinkedIdentityField,
    DrillDownHyperlinkedRelatedField,
)
from oscarapi.serializers.utils import (
    OscarModelSerializer,
    OscarHyperlinkedModelSerializer,
)
from oscarapi.serializers.fields import TaxIncludedDecimalField

logger = logging.getLogger(__name__)

User = get_user_model()
Basket = get_model("basket", "Basket")
Line = get_model("basket", "Line")
LineAttribute = get_model("basket", "LineAttribute")
StockRecord = get_model("partner", "StockRecord")
Voucher = get_model("voucher", "Voucher")
Product = get_model("catalogue", "Product")


class VoucherSerializer(OscarModelSerializer):
    class Meta:
        model = Voucher
        fields = overridable(
            "OSCARAPI_VOUCHER_FIELDS",
            default=("name", "code", "start_datetime", "end_datetime"),
        )


class OfferDiscountSerializer(
    serializers.Serializer
):  # pylint: disable=abstract-method
    description = serializers.CharField()
    name = serializers.CharField()
    amount = serializers.DecimalField(
        decimal_places=2, max_digits=12, source="discount"
    )


class VoucherDiscountSerializer(OfferDiscountSerializer):
    voucher = VoucherSerializer(required=False)


class BasketSerializer(serializers.HyperlinkedModelSerializer):
    lines = serializers.HyperlinkedIdentityField(view_name="basket-lines-list")
    offer_discounts = OfferDiscountSerializer(many=True, required=False)
    total_excl_tax = serializers.DecimalField(
        decimal_places=2, max_digits=12, required=False
    )
    total_excl_tax_excl_discounts = serializers.DecimalField(
        decimal_places=2, max_digits=12, required=False
    )
    total_incl_tax = TaxIncludedDecimalField(
        excl_tax_field="total_excl_tax", decimal_places=2, max_digits=12, required=False
    )
    total_incl_tax_excl_discounts = TaxIncludedDecimalField(
        excl_tax_field="total_excl_tax_excl_discounts",
        decimal_places=2,
        max_digits=12,
        required=False,
    )
    total_tax = TaxIncludedDecimalField(
        excl_tax_value=Decimal("0.00"), decimal_places=2, max_digits=12, required=False
    )
    currency = serializers.CharField(required=False)
    voucher_discounts = VoucherDiscountSerializer(many=True, required=False)
    owner = serializers.HyperlinkedRelatedField(
        view_name="user-detail",
        required=False,
        allow_null=True,
        queryset=User.objects.all(),
    )

    class Meta:
        model = Basket
        fields = overridable(
            "OSCARAPI_BASKET_FIELDS",
            default=(
                "id",
                "owner",
                "status",
                "lines",
                "url",
                "total_excl_tax",
                "total_excl_tax_excl_discounts",
                "total_incl_tax",
                "total_incl_tax_excl_discounts",
                "total_tax",
                "currency",
                "voucher_discounts",
                "offer_discounts",
                "is_tax_known",
            ),
        )


class LineAttributeSerializer(OscarHyperlinkedModelSerializer):
    url = DrillDownHyperlinkedIdentityField(
        view_name="lineattribute-detail",
        extra_url_kwargs={"basket_pk": "line.basket.id", "line_pk": "line.id"},
    )
    line = DrillDownHyperlinkedIdentityField(
        view_name="basket-line-detail", extra_url_kwargs={"basket_pk": "line.basket.id"}
    )

    class Meta:
        model = LineAttribute
        fields = "__all__"


class BasketLineSerializer(OscarHyperlinkedModelSerializer):
    """
    This serializer computes the prices of this line by using the basket
    strategy.
    """

    url = DrillDownHyperlinkedIdentityField(
        view_name="basket-line-detail", extra_url_kwargs={"basket_pk": "basket.id"}
    )
    attributes = LineAttributeSerializer(
        many=True, fields=("url", "option", "value"), required=False, read_only=True
    )
    price_excl_tax = serializers.DecimalField(
        decimal_places=2,
        max_digits=12,
        source="line_price_excl_tax_incl_discounts",
        read_only=True,
    )
    price_incl_tax = TaxIncludedDecimalField(
        decimal_places=2,
        max_digits=12,
        excl_tax_field="line_price_excl_tax_incl_discounts",
        source="line_price_incl_tax_incl_discounts",
        read_only=True,
    )
    price_incl_tax_excl_discounts = TaxIncludedDecimalField(
        decimal_places=2,
        max_digits=12,
        excl_tax_field="line_price_excl_tax",
        source="line_price_incl_tax",
        read_only=True,
    )
    price_excl_tax_excl_discounts = serializers.DecimalField(
        decimal_places=2, max_digits=12, source="line_price_excl_tax", read_only=True
    )
    warning = serializers.CharField(
        read_only=True, required=False, source="get_warning"
    )

    stockrecord = DrillDownHyperlinkedRelatedField(
        view_name="product-stockrecord-detail",
        extra_url_kwargs={"product_pk": "product_id"},
        queryset=StockRecord.objects.all(),
    )

    class Meta:
        model = Line
        fields = overridable(
            "OSCARAPI_BASKETLINE_FIELDS",
            default=(
                "url",
                "product",
                "quantity",
                "attributes",
                "price_currency",
                "price_excl_tax",
                "price_incl_tax",
                "price_incl_tax_excl_discounts",
                "price_excl_tax_excl_discounts",
                "is_tax_known",
                "warning",
                "basket",
                "stockrecord",
                "date_created",
                "date_updated",
            ),
        )

    def to_representation(self, obj):
        # This override is needed to reflect offer discounts or strategy
        # related prices immediately in the response
        operations.assign_basket_strategy(obj.basket, self.context["request"])

        # Oscar stores the calculated discount in line._discount_incl_tax or
        # line._discount_excl_tax when offers are applied. So by just
        # retrieving the line from the db you will loose this values, that's
        # why we need to get the line from the in-memory resultset here
        lines = (x for x in obj.basket.all_lines() if x.id == obj.id)
        line = next(lines, None)

        return super(BasketLineSerializer, self).to_representation(line)


class VoucherAddSerializer(serializers.Serializer):
    vouchercode = serializers.CharField(max_length=128, required=True)

    def validate(self, attrs):
        # oscar expects this always to be uppercase.
        attrs["vouchercode"] = attrs["vouchercode"].upper()

        request = self.context.get("request")
        try:
            voucher = Voucher.objects.get(code=attrs.get("vouchercode"))

            # check expiry date
            if not voucher.is_active():
                message = _("The '%(code)s' voucher has expired") % {
                    "code": voucher.code
                }
                raise serializers.ValidationError(message)

            # check voucher rules
            is_available, message = voucher.is_available_to_user(request.user)
            if not is_available:
                raise serializers.ValidationError(message)
        except Voucher.DoesNotExist:
            raise serializers.ValidationError(_("Voucher code unknown"))

        # set instance to the voucher so we can use this in the view
        self.instance = voucher
        return attrs

    def create(self, validated_data):
        return Voucher.objects.create(**validated_data)



import collections

from django.conf import settings

from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse


__all__ = ("api_root",)


def PUBLIC_APIS(r, f):
    return [
        ("register", reverse("api-register", request=r, format=f)),
        ("login", reverse("api-login", request=r, format=f)),
        ("basket", reverse("api-basket", request=r, format=f)),
        ("basket-add-product", reverse("api-basket-add-product", request=r, format=f)),
        ("basket-add-voucher", reverse("api-basket-add-voucher", request=r, format=f)),
        (
            "basket-shipping-methods",
            reverse("api-basket-shipping-methods", request=r, format=f),
        ),
        ("baskets", reverse("basket-list", request=r, format=f)),
        ("categories", reverse("category-list", request=r, format=f)),
        ("checkout", reverse("api-checkout", request=r, format=f)),
        ("orders", reverse("order-list", request=r, format=f)),
        ("options", reverse("option-list", request=r, format=f)),
        ("products", reverse("product-list", request=r, format=f)),
        ("countries", reverse("country-list", request=r, format=f)),
        ("useraddresses", reverse("useraddress-list", request=r, format=f)),
    ]


def ADMIN_APIS(r, f):
    return [
        ("productclasses", reverse("admin-productclass-list", request=r, format=f)),
        ("products", reverse("admin-product-list", request=r, format=f)),
        ("categories", reverse("admin-category-list", request=r, format=f)),
        ("orders", reverse("admin-order-list", request=r, format=f)),
        ("partners", reverse("admin-partner-list", request=r, format=f)),
        ("users", reverse("admin-user-list", request=r, format=f)),
    ]


@api_view(("GET",))
def api_root(
    request, format=None, *args, **kwargs
):  # pylint: disable=redefined-builtin
    """
    GET:
    Display all available urls.

    Since some urls have specific permissions, you might not be able to access
    them all.
    """
    apis = PUBLIC_APIS(request, format)

    if (
        not getattr(settings, "OSCARAPI_BLOCK_ADMIN_API_ACCESS", True)
        and request.user.is_staff
    ):
        apis += [("admin", collections.OrderedDict(ADMIN_APIS(request, format)))]

    return Response(collections.OrderedDict(apis))
